"""Custom error classes for USFM grammar related exceptions"""

class USFMGrammarError(Exception):
    """Custom error class for USFM grammar related exceptions"""

class ParsingError(Exception):
    """Custom error class for errors during parsing of USFM/USX/USJ data"""

class ParameterError(Exception):
    """Custom error class for errors related to input data or parameters"""



"""Convert other formats back into USFM"""

import re
from usfm_grammar.errors import ParsingError, USFMGrammarError

NO_USFM_USJ_TYPES = ["USJ", "table"]
NO_NEWLINE_USJ_TYPES = ["char", "note", "verse", "table:cell"]
CLOSING_USJ_TYPES = ["char", "note", "figure", "ref"]
NON_ATTRIB_USJ_KEYS = [
    "type",
    "marker",
    "content",
    "number",
    "sid",
    "code",
    "caller",
    "align",
    "version",
    "altnumber",
    "pubnumber",
    "category",
]
NON_ATTRIB_USX_KEYS = [
    "number",
    "code",
    "caller",
    "align",
    "sid",
    "eid",
    "style",
    "closed",
    "vid",
    "status",
    "version",
    "altnumber",
    "pubnumber",
    "category",
]
NO_NEWLINE_USX_TYPES = ["char", "note", "cell", "figure", "usx", "book", "optbreak"]
CLOSING_USX_TYPES = ["char", "note", "figure", "ms"]


class USFMGenerator:
    """Combines the different methods that generate USFM from other formats in one class"""

    def __init__(self):
        self.usfm_string = ""
        self.warnings = []

    # def is_valid_usfm(self, usfm_string: str = None) -> bool:
    #     '''Check the generated or passed USFM's correctness using the grammar'''
    #     if usfm_string is None:
    #         usfm_string = self.usfm_string
    #     return False

    def usj_to_usfm(self, usj_obj: dict, nested=False) -> None:  # pylint: disable=too-many-statements, too-many-branches
        """Traverses through the dict/json and uses 'type' field to form USFM elements"""
        if not isinstance(usj_obj, dict) or "type" not in usj_obj:
            raise USFMGrammarError("Unable to do the conversion. Ensure USJ is valid!")
        if usj_obj["type"] == "optbreak":
            if self.usfm_string != "" and self.usfm_string[-1] not in [
                "\n",
                "\r",
                " ",
                "\t",
            ]:
                self.usfm_string += " "
            self.usfm_string += "// "
            return
        if usj_obj["type"] == "ref":
            usj_obj["marker"] = "ref"
        if usj_obj["type"] not in NO_USFM_USJ_TYPES:
            self.usfm_string += "\\"
            if (
                nested
                and usj_obj["type"] == "char"
                and usj_obj["marker"] not in ["xt", "fv", "ref"]
            ):
                self.usfm_string += "+"
            self.usfm_string += f"{usj_obj['marker']} "
        if "code" in usj_obj:
            self.usfm_string += f"{usj_obj['code']} "
        if "number" in usj_obj:
            self.usfm_string += usj_obj["number"]
            if usj_obj["type"] == "verse":
                self.usfm_string += " "
        if "caller" in usj_obj:
            self.usfm_string += f"{usj_obj['caller']} "
        if "category" in usj_obj:
            self.usfm_string += f"\\cat {usj_obj['category']}\\cat*\n"
        if "content" in usj_obj:
            for item in usj_obj["content"]:
                if isinstance(item, str):
                    self.usfm_string += item
                else:
                    if usj_obj["type"] in ["char"]:
                        self.usj_to_usfm(item, nested=True)
                    else:
                        self.usj_to_usfm(item)
        attributes = False
        for key in usj_obj:
            if key not in NON_ATTRIB_USJ_KEYS:
                if not attributes:
                    self.usfm_string += "|"
                    attributes = True
                if key == "file":
                    self.usfm_string += f'src="{usj_obj[key]}" '
                else:
                    self.usfm_string += f'{key}="{usj_obj[key]}" '

        if usj_obj["type"] in CLOSING_USJ_TYPES:
            self.usfm_string = self.usfm_string.strip() + " \\"
            if (
                nested
                and usj_obj["type"] == "char"
                and usj_obj["marker"] not in ["xt", "ref", "fv"]
            ):
                self.usfm_string += "+"
            self.usfm_string += f"{usj_obj['marker']}* "
        if usj_obj["type"] == "ms":
            if "sid" in usj_obj:
                if not attributes:
                    self.usfm_string += "|"
                    attributes = True
                self.usfm_string += f'sid="{usj_obj["sid"]}" '
            self.usfm_string = self.usfm_string.strip() + "\\*"
        if usj_obj["type"] == "sidebar":
            self.usfm_string += "\\esbe"
        if usj_obj["type"] not in NO_NEWLINE_USJ_TYPES and self.usfm_string[-1] != "\n":
            self.usfm_string += "\n"
        if "altnumber" in usj_obj:
            self.usfm_string += f"\\{usj_obj['marker']}a {usj_obj['altnumber']}"
            self.usfm_string += f"\\{usj_obj['marker']}a* "
        if "pubnumber" in usj_obj:
            self.usfm_string += f"\\{usj_obj['marker']}p {usj_obj['pubnumber']}"
            if usj_obj["marker"] == "v":
                self.usfm_string += f"\\{usj_obj['marker']}p* "
            else:
                self.usfm_string += "\n"

    def usx_to_usfm(self, xml_obj, nested=False):  # pylint: disable=too-many-statements, too-many-branches
        """Traverses xml tree and converts nodes to usfm elements
        based on type and style fields"""
        if isinstance(xml_obj, str):
            self.usfm_string += xml_obj
            return
        obj_type = xml_obj.tag
        marker = None
        usfm_attributes = []
        if obj_type in ["verse", "chapter"] and "eid" in xml_obj.attrib:
            return
        if obj_type not in NO_NEWLINE_USX_TYPES:
            self.usfm_string += "\n"
        if obj_type == "optbreak":
            if self.usfm_string != "" and self.usfm_string[-1] not in [
                "\n",
                "\r",
                " ",
                "\t",
            ]:
                self.usfm_string += " "
            self.usfm_string += "// "
        if "style" in xml_obj.attrib:
            marker = xml_obj.attrib["style"]
            if nested and obj_type == "char" and marker not in ["xt", "fv", "ref"]:
                marker = "+" + marker
            self.usfm_string += f"\\{marker} "
        if "code" in xml_obj.attrib:
            self.usfm_string += xml_obj.attrib["code"]
        if "number" in xml_obj.attrib:
            self.usfm_string += f"{xml_obj.attrib['number']} "
        if "caller" in xml_obj.attrib:
            self.usfm_string += f"{xml_obj.attrib['caller']} "
        if "altnumber" in xml_obj.attrib:
            if obj_type == "verse":
                self.usfm_string += f"\\va {xml_obj.attrib['altnumber']}\\va*"
            elif obj_type == "chapter":
                self.usfm_string += f"\n\\ca {xml_obj.attrib['altnumber']}\\ca*"
        if "pubnumber" in xml_obj.attrib:
            if obj_type == "verse":
                self.usfm_string += f"\\vp {xml_obj.attrib['pubnumber']}\\vp*"
            elif obj_type == "chapter":
                self.usfm_string += f"\n\\cp {xml_obj.attrib['pubnumber']}"
        if "category" in xml_obj.attrib:
            self.usfm_string += f"\n\\cat {xml_obj.attrib['category']} \\cat*"
        if xml_obj.text:
            if self.usfm_string != "" and self.usfm_string[-1] not in [
                "\n",
                "\r",
                " ",
                "\t",
            ]:
                self.usfm_string += " "
            self.usfm_string += xml_obj.text.strip()
        for child in xml_obj.getchildren():
            if obj_type in ["char"]:
                self.usx_to_usfm(child, nested=True)
            else:
                self.usx_to_usfm(child, nested=False)
            if child.tail:
                if self.usfm_string != "" and self.usfm_string[-1] not in [
                    "\n",
                    "\r",
                    " ",
                    "\t",
                ]:
                    self.usfm_string += " "
                self.usfm_string += child.tail.strip()
        for key in xml_obj.attrib:
            val = xml_obj.attrib[key]
            val = val.replace('"', "")
            if key == "file" and obj_type == "figure":
                usfm_attributes.append(f'src="{val}"')
            elif key not in NON_ATTRIB_USX_KEYS:
                usfm_attributes.append(f'{key}="{val}"')
            if key in ["sid", "eid"] and obj_type == "ms":
                usfm_attributes.append(f'{key}="{val}"')
        if len(usfm_attributes) > 0:
            self.usfm_string += "|"
            self.usfm_string += " ".join(usfm_attributes)

        if (
            ("closed" in xml_obj.attrib and xml_obj.attrib["closed"] == "true")
            or obj_type in CLOSING_USX_TYPES
            or len(usfm_attributes) > 0
        ):
            # if not ("closed" in xml_obj.attrib and xml_obj.attrib['closed']=="false"):
            if obj_type == "ms":
                self.usfm_string += "\\*"
            else:
                self.usfm_string += f"\\{marker}*"
        if obj_type == "sidebar":
            self.usfm_string += "\n\\esbe\n"

    def biblenlp_to_usfm(self, biblenlp: dict, book_code: str = None) -> None:
        """Traverses through the verse texts and vrefs to generate a minimal USFM from it"""
        curr_book = None
        curr_chapter = None
        vref_pattern = re.compile(r"(\w\w\w) (\d+):(.*)")

        if (
            "text" not in biblenlp
            or "vref" not in biblenlp
            or not isinstance(biblenlp["vref"], list)
            or not isinstance(biblenlp["text"], list)
        ):
            raise ParsingError(
                "Incorrect format: "
                + "BibleNlp object should contain a dict with 'vref' and 'text' lists."
            )
        vrefs = biblenlp["vref"]
        if len(biblenlp["text"]) in [31170, 23213] and len(vrefs) == 41899:
            vrefs = vrefs[: len(biblenlp["text"])]
            biblenlp["vref"] = vrefs
        if book_code:
            book_code = book_code.strip().upper()
            vrefs = [
                ref.strip().upper()
                for ref in biblenlp["vref"]
                if ref.strip().upper().startswith(book_code)
            ]
        if len(vrefs) != len(biblenlp["text"]):
            if len(biblenlp["vref"]) == len(biblenlp["text"]) and book_code:
                texts = [
                    txt
                    for txt, ref in zip(biblenlp["text"], biblenlp["vref"])
                    if ref.strip().upper().startswith(book_code)
                ]
                biblenlp["text"] = texts
            if len(vrefs) != len(biblenlp["text"]):
                raise ParsingError(
                    "Incorrect format: Missmatch in lengths of vref and text lists."
                    + "Specify a book_code or check for versification differences. "
                    + f"{len(vrefs)} != {len(biblenlp['text'])}"
                )

        for vref, versetext in zip(vrefs, biblenlp["text"]):
            ref_match = re.match(vref_pattern, vref)
            if ref_match is None:
                raise ParsingError(
                    f"Incorrect format: {vref}.\nIn BibleNlp, vref should have "
                    + "three letter book code, chapter and verse in the following format: GEN 1:1"
                )
            book = ref_match.group(1)
            book = book.upper()
            chap = ref_match.group(2)
            verse = ref_match.group(3)
            if book != curr_book:
                if curr_book is not None:
                    self.warnings.append(
                        "USFM can contain only one book per file. "
                        + f"Only {curr_book} is processed. Specify book_code for other books."
                    )
                    break
                self.usfm_string += f"\\id {book}"
                curr_book = book
            if chap != curr_chapter:
                self.usfm_string += f"\n\\c {chap}\n\\p\n"
                curr_chapter = chap
            if not self.usfm_string.endswith("\n"):
                self.usfm_string += " "
            self.usfm_string += f"\\v {verse} {versetext}"


if __name__ == "__main__":
    from lxml import etree

    TEST_FILE = "../../../tests/basic/cross-refs/origin.xml"

    with open(TEST_FILE, "r", encoding="utf-8") as usx_file:
        usx_str = usx_file.read()
        root = etree.fromstring(usx_str)

        gen = USFMGenerator()

        gen.usx_to_usfm(root)
        print(gen.usfm_string)



"""The core logics of converting the syntax tree to other formats"""

from enum import Enum
import re
import traceback
import tree_sitter_usfm3 as tsusfm
from tree_sitter import Language, Parser, Query, QueryCursor
from lxml import etree
from usfm_grammar.usx_generator import USXGenerator
from usfm_grammar.usj_generator import USJGenerator
from usfm_grammar.list_generator import ListGenerator
from usfm_grammar.usfm_generator import USFMGenerator
from usfm_grammar.filters import exclude_markers_in_usj, include_markers_in_usj
from usfm_grammar.errors import USFMGrammarError, ParsingError, ParameterError

class Filter(list, Enum):
    """Defines the values of filter options"""

    BOOK_HEADERS = [
        "ide", "usfm", "h", "toc", "toca",  # identification
        "imt", "is", "ip", "ipi", "im", "imi", "ipq", "imq", "ipr",
        "iq", "ib", "ili", "iot", "io", "iex", "imte", "ie",  # intro
    ]
    TITLES = [
        "mt", "mte", "cl", "cd", "ms", "mr", "s", "sr", "r", "d", "sp", "sd",
    ]  # "headings"
    COMMENTS = ["sts", "rem", "lit", "restore"]  # comment markers
    PARAGRAPHS = [
        "p", "m", "po", "pr", "cls", "pmo", "pm", "pmc",  #'paragraphs-quotes-lists-tables'
        "pmr", "pi", "mi", "nb", "pc", "ph", "q", "qr", "qc", "qa", "qm", "qd",
        "lh", "li", "lf", "lim", "litl", "tr", "tc", "th", "tcr", "thr", "table", "b",
    ]
    CHARACTERS = [
        "add", "bk", "dc", "ior", "iqt", "k", "litl", "nd", "ord", "pn", "png",
        "qac", "qs", "qt", "rq", "sig", "sls", "tl", "wj",  # Special-text
        "em", "bd", "bdit", "it", "no", "sc", "sup",  # character styling
        "rb", "pro", "w", "wh", "wa", "wg",  # special-features
        "lik", "liv",  # structred list entries
        "jmp",
    ]
    NOTES = [
        "f", "fe", "ef", "efe", "x", "ex",  # "footnotes-and-crossrefs"
        "fr", "ft", "fk", "fq", "fqa", "fl", "fw", "fp", "fv", "fdc", "xo",
        "xop", "xt", "xta", "xk", "xq", "xot", "xnt", "xdc",
    ]
    STUDY_BIBLE = ["esb", "cat"]  # "sidebars-extended-contents"
    BCV = ["id", "c", "v"]
    TEXT = ["text-in-excluded-parent", "text"]
    # INNER_CONTENT = ['content-in-excluded-parent']


class Format(str, Enum):
    """Defines the valid values for input and output formats"""

    JSON = "usj"
    CSV = "table"
    ST = "syntax-tree"
    USX = "usx"
    MD = "markdown"
    USFM = "usfm"
    BIBLENLP = "biblenlp"


USFM_LANGUAGE = Language(tsusfm.language())
parser = Parser(USFM_LANGUAGE)


# # Handled alike by the node_2_dict_generic method
# ANY_VALID_MARKER = PARA_STYLE_MARKERS+NOTE_MARKERS+CHAR_STYLE_MARKERS+\
# NESTED_CHAR_STYLE_MARKERS+TABLE_CELL_MARKERS+\
# MISC_MARKERS


id_query = Query(USFM_LANGUAGE,
    """(book (id (bookcode) @book-code (description) @desc))"""
)

chapter_data_query = Query(USFM_LANGUAGE, """(c (chapterNumber) @chapter-number)
                                            (cl (text) @cl-text)
                                            (cp (text) @cp-text)
                                            (ca (chapterNumber) @ca-number)
                                            (cd) @cd-node""")

verse_data_query = Query(USFM_LANGUAGE, """(v (verseNumber) @verse-number)
                                            (vp (text) @vp-text)
                                            (va (verseNumber) @va-number)""")
error_query = Query(USFM_LANGUAGE, """(ERROR) @errors""")


class USFMParser:
    """Parser class with usfmstring, syntax_tree and methods for JSON convertions"""

    def __init__(  # pylint: disable=too-many-arguments, too-many-branches, too-many-positional-arguments
        self,
        usfm_string: str = None,
        from_usj: dict = None,
        from_usx: etree.Element = None,
        from_biblenlp: dict = None,
        book_code: str = None,
    ):
        # super(USFMParser, self).__init__()
        self.usfm_bytes = None
        self.syntax_tree = None
        self.errors = []
        self.warnings = []

        inputs_given = 0
        if usfm_string is not None:
            inputs_given += 1
        if from_usj is not None:
            inputs_given += 1
        if from_usx is not None:
            inputs_given += 1
        if from_biblenlp is not None:
            inputs_given += 1

        if inputs_given > 1:
            raise ParameterError(
                "Found more than one input!"
                + " Only one of USFM, USJ, USX or BibleNlp is supported in one object."
            )
        if inputs_given == 0:
            raise ParameterError(
                "Missing input! Either USFM, USJ, USX or BibleNlp is to be provided."
            )

        if usfm_string is not None:
            if not usfm_string.strip().startswith("\\"):
                raise ParameterError(
                    "Invalid input for USFM. Expected a string with \\ markups."
                )
            self.usfm = usfm_string
        elif from_usj is not None:
            usj_converter = USFMGenerator()
            usj_converter.usj_to_usfm(from_usj)
            self.usfm = usj_converter.usfm_string
        elif from_usx is not None:
            usx_converter = USFMGenerator()
            usx_converter.usx_to_usfm(from_usx)
            self.usfm = usx_converter.usfm_string
        elif from_biblenlp is not None:
            biblenlp_converter = USFMGenerator()
            biblenlp_converter.biblenlp_to_usfm(from_biblenlp, book_code)
            self.usfm = biblenlp_converter.usfm_string
            self.warnings.extend(biblenlp_converter.warnings)

        # Some basic sanity checks
        lower_case_book_code = re.compile(r"^\\id ([a-z0-9][a-z][a-z])")
        if re.match(lower_case_book_code, self.usfm):
            self.warnings.append("Found Book Code in lower case")
            found_book_code = re.match(lower_case_book_code, self.usfm).group(1)
            upper_book_code = found_book_code.upper()
            self.usfm = self.usfm.replace(found_book_code, upper_book_code, 1)

        self.usfm_bytes = bytes(self.usfm, "utf8")
        self.syntax_tree = parser.parse(self.usfm_bytes).root_node

        # check for errors in the parse tree and raise them
        err_cursor = QueryCursor(error_query)
        error_caps = err_cursor.captures(self.syntax_tree)
        if 'errors' in error_caps and len(error_caps['errors']) > 0:
            self.errors = [
                (
                    f"At {err.start_point}",
                    self.usfm_bytes[err.start_byte : err.end_byte].decode(
                        "utf-8"
                    ),
                )
                for err in error_caps["errors"]
            ]
        self.check_for_missing(self.syntax_tree)

    def check_for_missing(self, node):
        """Identify and report the MISSING nodes also as errors"""
        for child in node.children:
            if child.is_missing:
                self.errors.append((f"At {child.start_point}", f"Missing {child.type}"))
            else:
                self.check_for_missing(child)

    def to_syntax_tree(self, ignore_errors=False):
        """gives the syntax tree from class, as a string"""
        if not ignore_errors and self.errors:
            err_str = "\n\t".join([":".join(err) for err in self.errors])
            raise ParsingError(
                "Errors present:"
                + f"\n\t{err_str}"
                + "\nUse ignore_errors=True, to generate output inspite of errors"
            )
        return str(self.syntax_tree)  # Changed from .sexp() to str()

    def to_usj(
        self,
        exclude_markers=None,
        include_markers=None,
        ignore_errors=False,
        combine_texts=True,
    ):
        """convert the syntax_tree to the JSON format USJ.
        Filtering of desired contents using exclude markers option"""
        if not ignore_errors and self.errors:
            err_str = "\n\t".join([":".join(err) for err in self.errors])
            raise ParsingError(
                "Errors present:"
                + f"\n\t{err_str}"
                + "\nUse ignore_errors=True, to generate output inspite of errors"
            )
        json_root_obj = {"type": "USJ", "version": "3.1", "content": []}
        try:
            usj_generator = USJGenerator(USFM_LANGUAGE, self.usfm_bytes, json_root_obj)
            usj_generator.get_usj(self.syntax_tree, json_root_obj)
        except Exception as exe:
            traceback.print_exc()
            message = "Unable to do the conversion. "
            if self.errors:
                err_str = "\n\t".join([":".join(err) for err in self.errors])
                message += f"Could be due to an error in the USFM\n\t{err_str}"
            raise USFMGrammarError(message) from exe
        output_usj = usj_generator.json_root_obj
        if include_markers:
            output_usj = include_markers_in_usj(
                output_usj, include_markers + ["USJ"], combine_texts
            )
        if exclude_markers:
            output_usj = exclude_markers_in_usj(
                output_usj, exclude_markers, combine_texts
            )
        return output_usj

    def to_list(
        self,
        exclude_markers=None,
        include_markers=None,
        ignore_errors=False,
        combine_texts=True,
    ):
        """uses the toJSON function and converts JSON to CSV
        To be re-implemented to work with the flat JSON schema"""
        if not ignore_errors and self.errors:
            err_str = "\n\t".join([":".join(err) for err in self.errors])
            raise ParsingError(
                "Errors present:"
                + f"\n\t{err_str}"
                + "\nUse ignore_errors=True, to generate output inspite of errors"
            )

        try:
            include_list = None
            exclude_list = None
            if include_markers:
                include_list = include_markers + ["USJ"] + Filter.BCV
            if exclude_markers:
                exclude_list = [
                    mrkr for mrkr in exclude_markers if mrkr not in Filter.BCV
                ]

            usj_dict = self.to_usj(
                exclude_markers=exclude_list,
                include_markers=include_list,
                ignore_errors=ignore_errors,
                combine_texts=combine_texts,
            )

            list_generator = ListGenerator()
            list_generator.usj_to_list(usj_dict, exclude_markers, include_markers)
        except Exception as exe:
            message = "Unable to do the conversion. "
            if self.errors:
                err_str = "\n\t".join([":".join(err) for err in self.errors])
                message += f"Could be due to an error in the USFM\n\t{err_str}"
            raise USFMGrammarError(message) from exe
        return list_generator.list

    def to_biblenlp_format(self, ignore_errors=False):
        """uses the toUSJ function with BCV and TEXT filters,
        and converts the JSON to lists of texts and vrefs."""
        if not ignore_errors and self.errors:
            err_str = "\n\t".join([":".join(err) for err in self.errors])
            raise ParsingError(
                "Errors present:"
                + f"\n\t{err_str}"
                + "\nUse ignore_errors=True, to generate output inspite of errors"
            )

        json_root_obj = {"type": "USJ", "version": "3.1", "content": []}
        try:
            usj_generator = USJGenerator(USFM_LANGUAGE, self.usfm_bytes, json_root_obj)
            usj_generator.node_2_usj(self.syntax_tree, json_root_obj)
            usj_dict = usj_generator.json_root_obj
            usj_dict = include_markers_in_usj(usj_dict, Filter.BCV + Filter.TEXT, True)

            list_generator = ListGenerator()
            list_generator.usj_to_biblenlp_format(usj_dict)
        except Exception as exe:
            message = "Unable to do the conversion. "
            if self.errors:
                err_str = "\n\t".join([":".join(err) for err in self.errors])
                message += f"Could be due to an error in the USFM\n\t{err_str}"
            raise USFMGrammarError(message) from exe
        return list_generator.bible_nlp_format

    def to_markdown(self):
        """query for chapter, paragraph, text structure"""
        return "yet to be implemeneted"

    def to_usx(self, ignore_errors=False):
        """convert the syntax_tree to the XML format USX"""
        if not ignore_errors and self.errors:
            err_str = "\n\t".join([":".join(err) for err in self.errors])
            raise ParsingError(
                "Errors present:"
                + f"\n\t{err_str}"
                + "\nUse ignore_errors=True, to generate output inspite of errors"
            )

        usx_root = etree.Element("usx")
        usx_root.set("version", "3.1")
        try:
            usx_generator = USXGenerator(USFM_LANGUAGE, self.usfm_bytes, usx_root)
            usx_generator.node_2_usx(self.syntax_tree, usx_root)
        except Exception as exe:
            message = "Unable to do the conversion. "
            if self.errors:
                err_str = "\n\t".join([":".join(err) for err in self.errors])
                message += f"Could be due to an error in the USFM\n\t{err_str}"
            raise USFMGrammarError(message) from exe
        return usx_generator.xml_root_node



"""Test the to_usx conversion API"""

from doctest import Example

import pytest
from lxml import etree
from lxml.doctestcompare import LXMLOutputChecker, PARSE_XML

from tests import (
    all_usfm_files,
    initialise_parser,
    negative_tests,
    find_all_markers,
    generate_USFM_from_USX,
    parse_USFM_string,
    exclude_USX_files,
)

lxml_object = etree.Element("Root")
checker = LXMLOutputChecker()

# with open("../schemas/usx.rnc", encoding='utf-8') as f:
#     usxrnc_doc  = f.read()
# relaxng = etree.RelaxNG.from_rnc_string(usxrnc_doc)

with open("../schemas/usx.rng", encoding="utf-8") as f:
    relaxng_doc = etree.parse(f)
relaxng = etree.RelaxNG(relaxng_doc)

test_files = all_usfm_files.copy()
for file in negative_tests:
    if file in test_files:
        test_files.remove(file)


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.timeout(30)
def test_successful_usx_converion(file_path):
    """Tests if input parses & converts to usx successfully"""
    test_parser = initialise_parser(file_path)
    assert not test_parser.errors, test_parser.errors
    usx_xml = test_parser.to_usx()
    assert isinstance(usx_xml, type(lxml_object)), test_parser.to_syntax_tree()


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.timeout(30)
def test_generated_usx_with_rnc_grammar(file_path):
    """Tests if input parses & converts to usx successfully and validates the usx against schema"""
    test_parser = initialise_parser(file_path)
    assert not test_parser.errors, test_parser.errors
    usx_xml = test_parser.to_usx()
    assert isinstance(usx_xml, type(lxml_object)), test_parser.to_syntax_tree()
    relaxng.assertValid(usx_xml)


good_testsuite_usxs = test_files.copy()
for file in [path.replace("xml", "usfm") for path in exclude_USX_files]:
    if file in good_testsuite_usxs:
        good_testsuite_usxs.remove(file)


@pytest.mark.parametrize("file_path", good_testsuite_usxs)
@pytest.mark.timeout(30)
def test_compare_usx_with_testsuite_samples(file_path):
    """Compare the generated USX with the origin.xml in test suite"""
    test_parser = initialise_parser(file_path)
    assert not test_parser.errors, test_parser.errors
    usx_xml = test_parser.to_usx()
    assert isinstance(usx_xml, type(lxml_object)), test_parser.to_syntax_tree()
    usx_file_path = file_path.replace("origin.usfm", "origin.xml")
    origin_xml = etree.parse(usx_file_path)
    message = checker.output_difference(
        Example("", etree.tostring(origin_xml).decode("utf-8")),
        etree.tostring(usx_xml),
        PARSE_XML,
    )
    assert checker.check_output(
        etree.tostring(origin_xml), etree.tostring(usx_xml), PARSE_XML
    ), message


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.timeout(30)
def test_testsuite_usx_with_rnc_grammar(file_path):
    """Check the origin.xml files in testsuite, with the rnc grammar"""
    usx_file_path = file_path.replace("origin.usfm", "origin.xml")
    origin_xml = etree.parse(usx_file_path)
    relaxng.assertValid(origin_xml)


def get_styles(element):
    """Recursive function to traverse all xml nodes and their style attributes"""
    styles = []
    if "style" in element.attrib:
        styles.append(element.attrib["style"])
    if element.tag in ["figure", "optbreak"]:
        styles.append(element.tag)
    if "altnumber" in element.attrib:
        styles.append("altnumber")
    if "pubnumber" in element.attrib:
        styles.append("pubnumber")
    if "category" in element.attrib:
        styles.append("category")
    if len(element) > 0:
        for child in element:
            styles += get_styles(child)
    return styles


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.timeout(30)
def test_all_markers_are_in_output(file_path):
    """Tests if all markers in USFM are present in output also"""
    test_parser = initialise_parser(file_path)
    assert not test_parser.errors, test_parser.errors

    all_markers_in_input = find_all_markers(file_path, keep_id=True, keep_number=True)

    usx_xml = test_parser.to_usx()
    all_styles = get_styles(usx_xml)
    replacements = {
        "cat": "category",
        "ca": "altnumber",
        "cp": "pubnumber",
        "va": "altnumber",
        "vp": "pubnumber",
        "b": "optbreak",
        "fig": "figure",
    }
    for marker in all_markers_in_input:
        synonym = marker
        if replacements.get(marker):
            synonym = replacements[marker]
        # if (marker in ["ts", "qt"] or marker.endswith("-s") or \
        #     marker.endswith("-e") or marker.startswith("z")):
        #     marker = "milestone"
        assert marker in all_styles or synonym in all_styles, marker


@pytest.mark.parametrize("file_path", good_testsuite_usxs)
@pytest.mark.timeout(30)
def test_usx_round_tripping(file_path):
    """Convert USFM to USJ and back to USFM.
    Compare first USFM and second USFM based on parse tree"""
    file_path = file_path.replace(".usfm", ".xml")
    if file_path in exclude_USX_files:
        return
    with open(file_path, "r", encoding="utf-8") as usx_file:
        usx_str = usx_file.read()
        usx_str = usx_str.replace('encoding="utf-8"', "")
        if 'status="invalid"' in usx_str:
            return
        usx_xml = etree.fromstring(usx_str)

        generated_USFM = generate_USFM_from_USX(usx_xml)
        test_parser2 = parse_USFM_string(generated_USFM)
        assert not test_parser2.errors, str(test_parser2.errors) + "\n" + generated_USFM

        # assert test_parser2.to_usx() == usx_xml, generated_USX not same as input USX



"""Test the to_dict or json conversion API"""

import json, os
import pytest
import re

from tests import all_usfm_files, initialise_parser, negative_tests
from tests import parse_USFM_string, generate_USFM_from_BibleNlp
from src.usfm_grammar import Filter, USFMParser

test_files = all_usfm_files.copy()
for file in negative_tests:
    if file in test_files:
        test_files.remove(file)


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.timeout(30)
def test_list_converions_without_filter(file_path):
    """Tests if input parses without errors"""
    test_parser = initialise_parser(file_path)
    assert not test_parser.errors, test_parser.errors
    usfm_list = test_parser.to_list()
    assert isinstance(usfm_list, list)


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.parametrize("exclude_markers", [["s", "r"]])
@pytest.mark.timeout(30)
def test_list_converions_with_exclude_markers(file_path, exclude_markers):
    """Tests if input parses without errors"""
    test_parser = initialise_parser(file_path)
    assert not test_parser.errors, test_parser.errors
    usfm_list = test_parser.to_list(exclude_markers=exclude_markers)
    assert isinstance(usfm_list, list)
    for row in usfm_list[1:]:
        assert row[5] not in exclude_markers


trailing_num_pattern = re.compile(r"\d+$")


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.parametrize(
    "include_markers", [["id", "c", "v"] + Filter.TEXT + Filter.PARAGRAPHS]
)
@pytest.mark.timeout(30)
def test_list_converions_with_include_markers(file_path, include_markers):
    """Tests if input parses without errors"""
    test_parser = initialise_parser(file_path)
    assert not test_parser.errors, test_parser.errors
    usfm_list = test_parser.to_list(include_markers=include_markers)
    assert isinstance(usfm_list, list)
    for row in usfm_list[1:]:
        marker = row[5]
        marker = re.sub(trailing_num_pattern, "", marker)
        assert marker in include_markers


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.timeout(30)
def test_usfm_to_biblenlp_conversion(file_path):
    """Tests if input parses without errors"""
    test_parser = initialise_parser(file_path)
    assert not test_parser.errors, test_parser.errors
    bible_nlp_dict = test_parser.to_biblenlp_format()
    assert isinstance(bible_nlp_dict, dict)
    assert "text" in bible_nlp_dict
    assert "vref" in bible_nlp_dict
    assert len(bible_nlp_dict["text"]) == len(bible_nlp_dict["vref"])


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.timeout(30)
def test_usj_to_biblenlp_conversion(file_path):
    """Tests if input parses without errors"""
    usj_path = file_path.replace("usfm", "json")
    if (
        os.path.isfile(usj_path)
        and "special-cases/empty-attributes/origin.json" not in usj_path
    ):
        with open(usj_path, "r", encoding="utf-8") as usj_fp:
            usj = json.load(usj_fp)
            test_parser = USFMParser(from_usj=usj)
            assert not test_parser.errors, test_parser.errors
            bible_nlp_dict = test_parser.to_biblenlp_format()
            assert isinstance(bible_nlp_dict, dict)
            assert "text" in bible_nlp_dict
            assert "vref" in bible_nlp_dict
            assert len(bible_nlp_dict["text"]) == len(bible_nlp_dict["vref"])


emtpy_scritures = [
    "../tests/mandatory/emptyV/origin.usfm",
    "../tests/mandatory/v/origin.usfm",
    "../tests/paratextTests/NoErrorsShort/origin.usfm",
    "../tests/special-cases/empty-book/origin.usfm",
]
for path in emtpy_scritures:
    test_files.remove(path)


@pytest.mark.parametrize("file_path", test_files)
@pytest.mark.timeout(30)
def test_biblenlp_to_usfm(file_path):
    """Test USFM to BibleNLP, then BibleNLP to USFM"""
    test_parser = initialise_parser(file_path)
    bible_nlp_dict = test_parser.to_biblenlp_format()

    generated_usfm = generate_USFM_from_BibleNlp(bible_nlp_dict)
    print(generated_usfm)

    test_parser2 = parse_USFM_string(generated_usfm)
    assert test_parser2.errors == []



"""
Tree sitter queries for Python USFM parser
"""
from tree_sitter import Query

def create_queries_as_needed(name, lang):
    """Create and return a query based on the given name and language.

    Args:
        name (str): The name of the query to create
        lang: The language object to create the query with

    Returns:
        The created query or None if name doesn't match any query
    """
    query_creators = {
        "chapter": get_chap_query,
        "usjCaVa": usj_ca_va_query,
        "attribVal": attrib_val_query,
        "para": para_query,
        "id": get_id_query,
        "milestone": milestone_query,
        "category": category_query,
        "verseNumCap": verse_num_cap_query
    }
    creator = query_creators.get(name)
    return creator(lang) if creator else None


def get_id_query(lang):
    """Create and return a query for book ID information."""
    return Query(lang, "(id (bookcode) @book-code (description)? @desc)")


def usj_ca_va_query(lang):
    """Create and return a query for chapter and verse alternate numbers."""
    return Query(lang,
        """
    ([
        (chapterNumber)
        (verseNumber)
    ] @alt-num)
    """.strip()
    )


def attrib_val_query(lang):
    """Create and return a query for attribute values."""
    return Query(lang, "((attributeValue) @attrib-val)")


def get_chap_query(lang):
    """Create and return a query for chapter information."""
    return Query(lang,
        """
    (c (chapterNumber) @chap-num
      (ca (chapterNumber) @alt-num)?
      (cp (text) @pub-num)?)
    """.strip()
    )


def para_query(lang):
    """Create and return a query for paragraph markers."""
    return Query(lang, "(paragraph (_) @para-marker)")


def milestone_query(lang):
    """Create and return a query for milestone tags."""
    return Query(lang,
        """
    ([
        (milestoneTag)
        (milestoneStartTag)
        (milestoneEndTag)
        (zSpaceTag)
    ] @ms-name)
    """.strip()
    )


def category_query(lang):
    """Create and return a query for categories."""
    return Query(lang, "((category) @category)")


def verse_num_cap_query(lang):
    """Create and return a query for verse number information."""
    return Query(lang,
        """
    (v
        (verseNumber) @vnum
        (va (verseNumber) @alt)?
        (vp (text) @vp)?
    )
    """.strip()
    )



"""Logics for syntax-tree to dict(USJ) conversions"""
from tree_sitter import QueryCursor

from usfm_grammar.queries import create_queries_as_needed
from usfm_grammar.usx_generator import USXGenerator

#pylint: disable=duplicate-code

class USJGenerator:
    """A binding for all methods used in generating USJ from Syntax tree"""
    MARKER_SETS = USXGenerator.MARKER_SETS
    MARKER_LISTS = USXGenerator.MARKER_LISTS
    DEFAULT_ATTRIB_MAP = USXGenerator.DEFAULT_ATTRIB_MAP

    def __init__(self, tree_sitter_language_obj, usfm_string, usj_root_obj=None):
        """Initialize the USJ generator with USFM and root object"""
        self.usfm_language = tree_sitter_language_obj
        self.usfm = usfm_string
        self.json_root_obj = usj_root_obj or {
            "type": "USJ",
            "version": "3.1",
            "content": [],
        }
        # Cache for the query objects
        self.queries = {}
        # Make o(1) sets for marker lookups
        self.parse_state = {"book_slug": None, "current_chapter": None}
        # maps and id to a fn;
        self.dispatch_map = self.populate_dispatch_map()

    def get_query(self, name):
        """Get or create a query by name"""
        if name not in self.queries:
            self.queries[name] = self.create_query(name)
        return self.queries[name]

    def create_query(self, name):
        """Create a query by name"""
        return create_queries_as_needed(name, self.usfm_language)

    def node_2_usj_id(self, node, parent_json_obj):
        """Convert ID node to USJ format"""
        id_cursor = QueryCursor(self.get_query("id"))
        id_captures = id_cursor.captures(node)
        code = self.usfm[id_captures['book-code'][0].start_byte :
                         id_captures['book-code'][0].end_byte].decode("utf-8")\
                    if 'book-code' in id_captures else None
        desc = self.usfm[id_captures['desc'][0].start_byte :
                         id_captures['desc'][0].end_byte].decode("utf-8")\
                    if 'desc' in id_captures else None

        book_json_obj = {"type": "book", "marker": "id", "code": code, "content": []}

        self.parse_state["book_slug"] = code
        if desc and desc.strip():
            book_json_obj["content"].append(desc.strip())
        parent_json_obj["content"].append(book_json_obj)

    def node_2_usj_c(self, node, parent_json_obj):
        """Convert chapter node to USJ format"""
        chap_cursor = QueryCursor(self.get_query("chapter"))
        chap_cap = chap_cursor.captures(node)

        chap_num = self.usfm[chap_cap['chap-num'][0].start_byte :
                             chap_cap['chap-num'][0].end_byte].decode("utf-8")\
            if 'chap-num' in chap_cap else None
        chap_ref = f"{self.parse_state['book_slug']} {chap_num}"

        chap_json_obj = {
            "type": "chapter",
            "marker": "c",
            "number": chap_num,
            "sid": chap_ref,
        }

        self.parse_state["current_chapter"] = chap_num

        alt_num = self.usfm[chap_cap['alt-num'][0].start_byte :
                             chap_cap['alt-num'][0].end_byte].decode("utf-8")\
            if 'alt-num' in chap_cap else None
        pub_num = self.usfm[chap_cap['pub-num'][0].start_byte :
                             chap_cap['pub-num'][0].end_byte].decode("utf-8")\
            if 'pub-num' in chap_cap else None

        if alt_num and alt_num.strip():
            chap_json_obj["altnumber"] = alt_num.strip()
        if pub_num and pub_num.strip():
            chap_json_obj["pubnumber"] = pub_num.strip()

        parent_json_obj["content"].append(chap_json_obj)

        for child in node.children:
            if child.type in ["cl", "cd"]:
                self.node_2_usj(child, parent_json_obj)

    def node_2_usj_chapter(self, node, parent_json_obj):
        """Convert chapter content to USJ format"""
        for child in node.children:
            if child.type == "c":
                self.node_2_usj_c(child, parent_json_obj)
            else:
                self.node_2_usj(child, parent_json_obj)

    def node_2_usj_verse(self, node, parent_json_obj):
        """Convert verse node to USJ format"""
        verse_cursor = QueryCursor(self.get_query("verseNumCap"))
        verse_num_cap = verse_cursor.captures(node)
        if 'vnum' not in verse_num_cap:
            return

        verse_node = verse_num_cap['vnum'][0]
        verse_num = (
            self.usfm[verse_node.start_byte : verse_node.end_byte]
            .decode("utf-8")
            .strip()
        )

        v_json_obj = {"type": "verse", "marker": "v", "number": verse_num}
        # Process additional verse attributes if present
        if 'alt' in verse_num_cap:
            alt_node = verse_num_cap['alt'][0]
            v_json_obj["altnumber"] = (
                self.usfm[alt_node.start_byte : alt_node.end_byte]
                .decode("utf-8")
                .strip()
            )
        if 'vp' in verse_num_cap:
            vp_node = verse_num_cap['vp'][0]
            v_json_obj["pubnumber"] = (
                self.usfm[vp_node.start_byte : vp_node.end_byte]
                .decode("utf-8")
                .strip()
            )

        # Add sid (scripture ID) using current book and chapter
        if self.parse_state.get("book_slug") and self.parse_state.get(
            "current_chapter"
        ):
            v_json_obj["sid"] = (
                f"{self.parse_state['book_slug']} {self.parse_state['current_chapter']}:{verse_num}"
            )
        parent_json_obj["content"].append(v_json_obj)

    def node_2_usj_ca_va(self, node, parent_json_obj):
        """Convert ca/va nodes to USJ format"""
        style = node.type
        char_json_obj = {"type": "char", "marker": style.strip()}

        alt_num_cursor = QueryCursor(self.get_query("usjCaVa"))
        alt_num_match = alt_num_cursor.captures(node)

        alt_num = self.usfm[
            alt_num_match["alt-num"][0].start_byte : alt_num_match["alt-num"][0].end_byte
        ].strip()

        char_json_obj["altnumber"] = alt_num
        parent_json_obj["content"].append(char_json_obj)

    def node_2_usj_para(self, node, parent_json_obj):
        """Convert paragraph nodes to USJ format"""
        if node.children and node.children[0].type.endswith("Block"):
            for child in node.children[0].children:
                self.node_2_usj_para(child, parent_json_obj)
        elif node.type == "paragraph":
            para_tag_cursor = QueryCursor(self.get_query("para"))
            para_tag_cap = para_tag_cursor.captures(node)
            if 'para-marker' not in para_tag_cap:
                return
            # Unpack the capture tuple
            para_node = para_tag_cap['para-marker'][0]
            para_marker = para_node.type
            if para_marker == "b":
                parent_json_obj["content"].append(
                    {"type": "para", "marker": para_marker}
                )
            elif not para_marker.endswith("Block"):
                para_json_obj = {"type": "para", "marker": para_marker, "content": []}
                for child in para_node.children:
                    self.node_2_usj(child, para_json_obj)
                parent_json_obj["content"].append(para_json_obj)
        elif node.type in ["pi", "ph"]:
            para_marker = (
                self.usfm[node.children[0].start_byte : node.children[0].end_byte]
                .decode("utf-8")
                .replace("\\", "")
                .strip()
            )
            para_json_obj = {"type": "para", "marker": para_marker, "content": []}
            for child in node.children[1:]:
                self.node_2_usj(child, para_json_obj)
            parent_json_obj["content"].append(para_json_obj)

    def node_2_usj_notes(self, node, parent_json_obj):
        """Convert footnote and cross-reference nodes to USJ format"""
        tag_node = node.children[0]
        caller_node = node.children[1]
        style = (
            self.usfm[tag_node.start_byte : tag_node.end_byte]
            .decode("utf-8")
            .replace("\\", "")
            .strip()
        )

        note_json_obj = {"type": "note", "marker": style.strip(), "content": []}

        note_json_obj["caller"] = (
            self.usfm[caller_node.start_byte : caller_node.end_byte]
            .decode("utf-8")
            .strip()
        )

        for i in range(2, len(node.children) - 1):
            self.node_2_usj(node.children[i], note_json_obj)

        parent_json_obj["content"].append(note_json_obj)

    def node_2_usj_char(self, node, parent_json_obj):
        """Convert character style nodes to USJ format"""
        tag_node = node.children[0]
        children_range = len(node.children)

        for i in range(len(node.children)-1, 0, -1):
            if node.children[i].type.startswith("\\") or\
                node.children[i].type == "*" or\
                node.children[i].type.endswith("Tag"):
                children_range -= 1

        style = (
            self.usfm[tag_node.start_byte : tag_node.end_byte]
            .decode("utf-8")
            .replace("\\", "")
            .replace("+", "")
            .strip()
        )
        char_json_obj = {"type": "char", "marker": style.strip(), "content": []}

        for i in range(1, children_range):
            self.node_2_usj(node.children[i], char_json_obj)

        parent_json_obj["content"].append(char_json_obj)

    def node_2_usj_table(self, node, parent_json_obj):
        """Convert table-related nodes to USJ format"""
        if node.type == "table":
            table_json_obj = {"type": "table", "content": []}
            for child in node.children:
                self.node_2_usj(child, table_json_obj)
            parent_json_obj["content"].append(table_json_obj)
        elif node.type == "tr":
            row_json_obj = {"type": "table:row", "marker": "tr", "content": []}
            for child in node.children[1:]:
                self.node_2_usj(child, row_json_obj)
            parent_json_obj["content"].append(row_json_obj)
        elif node.type in USJGenerator.MARKER_SETS["table_cell"]:
            tag_node = node.children[0]
            style = (
                self.usfm[tag_node.start_byte : tag_node.end_byte]
                .decode("utf-8")
                .replace("\\", "")
                .strip()
            )
            cell_json_obj = {
                "type": "table:cell",
                "marker": style.strip(),
                "content": [],
            }

            if "tcc" in style or "thc" in style:
                cell_json_obj["align"] = "center"
            elif style.endswith("r"):
                cell_json_obj["align"] = "end"
            else:
                cell_json_obj["align"] = "start"

            for child in node.children[1:]:
                self.node_2_usj(child, cell_json_obj)

            parent_json_obj["content"].append(cell_json_obj)

    def node_2_usj_attrib(self, node, parent_json_obj):
        """Add attribute values to USJ elements"""
        attrib_name_node = node.children[0]
        attrib_name = (
            self.usfm[attrib_name_node.start_byte : attrib_name_node.end_byte]
            .decode("utf-8")
            .strip()
        )

        if attrib_name == "|":
            parent_type = node.parent.type
            if "Nested" in parent_type:
                parent_type = parent_type.replace("Nested", "")
            attrib_name = USJGenerator.DEFAULT_ATTRIB_MAP.get(parent_type, attrib_name)

        if attrib_name == "src":
            attrib_name = "file"

        attrib_cursor = QueryCursor(self.get_query("attribVal"))
        attrib_val_cap = attrib_cursor.captures(node)
        attrib_value = self.usfm[attrib_val_cap['attrib-val'][0].start_byte :
                attrib_val_cap['attrib-val'][0].end_byte].decode('utf-8')\
            if "attrib-val" in attrib_val_cap else ""

        if attrib_val_cap:
            node_obj= attrib_val_cap['attrib-val'][0]
            attrib_value = (
                self.usfm[node_obj.start_byte : node_obj.end_byte]
                .decode("utf-8")
                .strip()
            )

        parent_json_obj[attrib_name] = attrib_value

    def node_2_usj_milestone(self, node, parent_json_obj):
        """Convert milestone nodes to USJ format"""
        ms_name_cursor = QueryCursor(self.get_query("milestone"))
        ms_name_cap = ms_name_cursor.captures(node)
        if 'ms-name' not in ms_name_cap:
            return

        # Unpack the first capture (node, capture_name)
        ms_node = ms_name_cap['ms-name'][0]
        style = (
            self.usfm[ms_node.start_byte : ms_node.end_byte]
            .decode("utf-8")
            .replace("\\", "")
            .strip()
        )
        ms_json_obj = {"type": "ms", "marker": style.strip(), "content": []}

        for child in node.children:
            if child.type.endswith("Attribute"):
                self.node_2_usj(child, ms_json_obj)

        if not ms_json_obj["content"]:
            del ms_json_obj["content"]

        parent_json_obj["content"].append(ms_json_obj)

    def node_2_usj_special(self, node, parent_json_obj):
        """Convert special nodes (esb, cat, fig, ref) to USJ format"""
        if node.type == "esb":
            sidebar_json_obj = {"type": "sidebar", "marker": "esb", "content": []}
            for child in node.children[1:-1]:
                self.node_2_usj(child, sidebar_json_obj)
            parent_json_obj["content"].append(sidebar_json_obj)
        elif node.type == "cat":
            cat_cursor = QueryCursor(self.get_query("category"))
            cat_cap = cat_cursor.captures(node)
            if 'category' not in cat_cap:
                return

            # Unpack the first capture (node, capture_name)
            cat_node = cat_cap['category'][0]
            category = (
                self.usfm[cat_node.start_byte : cat_node.end_byte]
                .decode("utf-8")
                .strip()
            )
            parent_json_obj["category"] = category
        elif node.type == "fig":
            fig_json_obj = {"type": "figure", "marker": "fig", "content": []}
            for child in node.children[1:-1]:
                self.node_2_usj(child, fig_json_obj)
            parent_json_obj["content"].append(fig_json_obj)
        elif node.type == "ref":
            ref_json_obj = {"type": "ref", "content": []}
            for child in node.children[1:-1]:
                self.node_2_usj(child, ref_json_obj)
            parent_json_obj["content"].append(ref_json_obj)

    def node_2_usj_generic(self, node, parent_json_obj):
        """Convert generic nodes to USJ format"""
        tag_node = node.children[0]
        style = self.usfm[tag_node.start_byte : tag_node.end_byte].decode("utf-8")
        if style.startswith("\\"):
            style = style[1:]
        else:
            style = node.type
        children_range_start = 1

        if len(node.children) > 1 and node.children[1].type.startswith("numbered"):
            num_node = node.children[1]
            num = self.usfm[num_node.start_byte : num_node.end_byte]
            style += num
            children_range_start = 2

        para_json_obj = {"type": "para", "marker": style.strip(), "content": []}
        parent_json_obj["content"].append(para_json_obj)

        for i in range(children_range_start, len(node.children)):
            child = node.children[i]
            if any(
                child.type in marker_set
                for marker_set in [
                    USJGenerator.MARKER_SETS["char_style"],
                    USJGenerator.MARKER_SETS["nested_char_style"],
                    USJGenerator.MARKER_SETS["other_para_nestables"],
                ]
            ):
                self.node_2_usj(child, para_json_obj)
            else:
                self.node_2_usj(child, parent_json_obj)

    def push_text_node(self, node, parent_json_obj):
        """Add text node to parent's content"""
        text_val = (
            self.usfm[node.start_byte : node.end_byte].decode("utf-8").replace("~", " ")
        )
        if text_val:
            parent_json_obj["content"].append(text_val)

    def handle_verse_text(self, node, parent_json_obj):
        """Handle verse text content"""
        for child in node.children:
            self.node_2_usj(child, parent_json_obj)

    def populate_dispatch_map(self):
        """Create a dispatch map for node type to handler functions"""
        dispatch_map = {}

        def add_handlers(markers, handler):
            for marker in markers:
                if isinstance(marker, (list, set)):
                    for mrkr in marker:
                        dispatch_map[mrkr] = getattr(self, handler.__name__)
                else:
                    dispatch_map[marker] = getattr(self, handler.__name__)

        # Add basic handlers
        dispatch_map["text"] = self.push_text_node
        dispatch_map["verseText"] = self.handle_verse_text
        dispatch_map["v"] = self.node_2_usj_verse
        dispatch_map["id"] = self.node_2_usj_id
        dispatch_map["chapter"] = self.node_2_usj_chapter
        dispatch_map["usfm"] = lambda *_: None  # noop

        # Add handlers for different marker types
        add_handlers(["paragraph", "q", "w"], self.node_2_usj_para)
        add_handlers(["cl", "cp", "vp"], self.node_2_usj_generic)
        add_handlers(["ca", "va"], self.node_2_usj_ca_va)
        add_handlers(["table", "tr"], self.node_2_usj_table)
        add_handlers(["milestone", "zNameSpace"], self.node_2_usj_milestone)
        add_handlers(["esb", "cat", "fig", "ref"], self.node_2_usj_special)
        add_handlers(USJGenerator.MARKER_LISTS["note"], self.node_2_usj_notes)
        add_handlers(
            USJGenerator.MARKER_LISTS["char_style"]
            + USJGenerator.MARKER_LISTS["nested_char_style"]
            + ["xt_standalone"],
            self.node_2_usj_char,
        )
        add_handlers(USJGenerator.MARKER_LISTS["table_cell"], self.node_2_usj_table)

        # Add paragraph style markers
        for marker in USJGenerator.MARKER_LISTS["para_style"]:
            if marker != "usfm":
                dispatch_map[marker] = self.node_2_usj_generic

        return dispatch_map

    def get_usj(self, node, parent_json_obj=None):
        """Convert the syntax tree to USJ format"""
        if parent_json_obj is None:
            parent_json_obj = self.json_root_obj

        self.node_2_usj(node, parent_json_obj)
        return parent_json_obj

    def node_2_usj(self, node, parent_json_obj):
        """Main dispatch method for converting nodes to USJ format"""
        if not hasattr(node, "type"):
            return
        node_type = node.type.replace("\\", "") if node.type else ""
        handler = self.dispatch_map.get(node_type)
        if handler:
            handler(node, parent_json_obj)
        elif node_type:
            # Handle special cases
            if node_type.endswith("Attribute"):
                self.node_2_usj_attrib(node, parent_json_obj)
            elif node_type.strip() in ["", "|"]:
                # Known noop
                return
            # Process children for nodes without specific handlers
            elif hasattr(node, "children"):
                for child in node.children:
                    self.node_2_usj(child, parent_json_obj)



"""Check the formats of USFM and USJ. Also tries to fixe common errors in USFM"""

import re
import jsonschema

import tree_sitter_usfm3 as tsusfm
from tree_sitter import Language, Parser, QueryCursor

from usfm_grammar.usfm_parser import error_query
from usfm_grammar.schema import usj_schema


class Validator:
    """Check validity of USJ and USFM. Also auto fix USFM"""

    def __init__(self, tree_sitter_usfm=tsusfm):
        """contrsuctor"""
        usfm_language = Language(tree_sitter_usfm.language())
        self.usfm_parser = Parser(usfm_language)
        self.usfm_errors = []

        self.usj_validator = jsonschema.validators.Draft7Validator(schema=usj_schema)

        self.message = ""
        self.modified_usfm = ""
        self.usfm_bytes = ""

    def is_valid_usj(self, usj):
        """Inputs: USJ, Output: True/false, details in message"""
        # validate(instance=usj, schema=self.usj_schema)
        self.message = ""
        try:
            self.usj_validator.validate(usj)
        except Exception as exce:  # pylint: disable=W0703
            self.message = str(exce)
            return False
        return True

    def is_valid_usfm(self, usfm):
        """Inputs: USFM, Output: True/false, details in message"""
        self.usfm_bytes = bytes(usfm, "utf8")
        tree = self.usfm_parser.parse(self.usfm_bytes)

        # check for errors in the parse tree and raise them
        err_cursor = QueryCursor(error_query)
        errors_cap = err_cursor.captures(tree.root_node)
        self.usfm_errors = []
        self.message = ""
        if 'errors' in errors_cap and len(errors_cap['errors']) > 0:
            self.usfm_errors = errors_cap['errors']
        self.check_for_missing(tree.root_node)

        if len(self.usfm_errors) > 0:
            self.message = self.format_errors()
            return False
        return True

    def check_for_missing(self, node):
        """Identify and report the MISSING nodes also as errors"""
        for child in node.children:
            if child.is_missing:
                self.usfm_errors.append(child)
            else:
                self.check_for_missing(child)

    def format_errors(self):
        """Prettify error messages from parse tree, to print"""
        err_lines = []
        for err in self.usfm_errors:
            # print("---"*10)
            # print(err.is_error)
            # print(err.is_missing)
            # print(err.text)
            # print(err.parent.sexp())
            # print(err.prev_sibling)
            # print(err.children[0].type)
            # print(f">>>{err.sexp()}<<<")
            # print("---"*10)
            if err.is_missing:
                start = max(0, err.start_byte - 3)
                end = min(len(self.usfm_bytes), err.start_byte + 10)
                err_lines.append(
                    (
                        f"At {err.start_point}:Missing",
                        self.usfm_bytes[start:end].decode("utf-8"),
                    )
                )
            else:
                err_lines.append(
                    (
                        f"At {err.start_point}",
                        self.usfm_bytes[err.start_byte : err.end_byte].decode("utf-8"),
                    )
                )
        err_str = "\n\t".join([":".join(err) for err in err_lines])
        err_str = f"Errors present:\n\t{err_str}"
        return err_str

    def auto_fix_usfm(self, usfm, fixed=False):  # pylint: disable=R0912,R0915
        """Iteratively tries to fix the errors found in usfm
        input: usfm string
        output: fixed usfm
        details in message"""
        if self.is_valid_usfm(usfm):
            self.message = "Fixed Errors in USFM" if fixed else "No Errors in USFM"
            return usfm
        self.modified_usfm = usfm
        changed = False
        for error in self.usfm_errors:
            error_text = error.text.decode("utf-8")
            # No \P after \s5
            if (
                error.is_error
                and error_text.startswith("\\s5")
                and "paragraph" not in [ch.type for ch in error.children]
            ):
                # print("match 1")
                self.modified_usfm = re.sub(
                    r"\\s5[\s\n\r]*", r"\\s5 \n\\p\n", self.modified_usfm
                )
                changed = True
            # Missing space after \s5
            elif (
                error.is_missing
                and error.parent.type == "sTag"
                and str(error) == '(MISSING " ")'
            ):
                # print("match 2")
                self.modified_usfm = re.sub(r"\\s5\n", r"\\s5 \n", self.modified_usfm)
                changed = True
            # book code is missing(empty id marker)
            elif re.match(book_code_missing_pattern, error_text):
                # print("match 3")
                self.modified_usfm = re.sub(r"\\id", r"\\id XXX", self.modified_usfm)
                changed = True
            # \p not given after section heading
            elif (
                error.is_error
                and error_text.startswith("\\v")
                and error.parent.type == "s"
                and "paragraph" not in [ch.type for ch in error.children]
            ):
                # print("match 4")
                start = error.parent.start_byte
                end = error.start_byte
                to_replace = self.usfm_bytes[start:end].decode("utf-8")
                repalcement = to_replace + "\\p\n"
                # self.modified_usfm = re.sub(to_replace, repalcement, self.modified_usfm)
                self.modified_usfm = self.modified_usfm.replace(to_replace, repalcement)
                changed = True
            # space missing between \v and number
            elif re.match(v_without_space_pattern, error_text):
                # print("match 5")
                self.modified_usfm = re.sub(
                    v_without_space_pattern, r"\1 \2", self.modified_usfm
                )
                changed = True
            # space missing between \c and number
            elif re.match(c_without_space_pattern, error_text):
                # print("match 6")
                self.modified_usfm = re.sub(
                    c_without_space_pattern, r"\1 \2", self.modified_usfm
                )
                changed = True
            # \p not given at chapter start
            elif (
                error.is_error
                and error_text.startswith("\\v")
                and error.prev_sibling.type == "chapter"
                and "paragraph" not in [ch.type for ch in error.children]
            ):
                # print("match 7")
                start = error.prev_sibling.start_byte
                end = error.start_byte
                to_replace = self.usfm_bytes[start:end].decode("utf-8")
                repalcement = to_replace + "\\p\n"
                # self.modified_usfm = re.sub(to_replace, repalcement, self.modified_usfm)
                self.modified_usfm = self.modified_usfm.replace(to_replace, repalcement)
                changed = True
            elif (
                error.is_error
                and (not error_text.startswith("\\"))
                and error.prev_sibling.type == "chapter"
                and "paragraph" not in [ch.type for ch in error.children]
            ):
                # print("match 7.1")
                start = error.prev_sibling.start_byte
                end = error.start_byte
                to_replace = self.usfm_bytes[start:end].decode("utf-8")
                repalcement = to_replace + "\\p\n"
                # self.modified_usfm = re.sub(to_replace, repalcement, self.modified_usfm)
                self.modified_usfm = self.modified_usfm.replace(to_replace, repalcement)
                changed = True
            # Stray slash not with a valid marker
            elif error_text.startswith("\\") and (
                not re.match(valid_markers_pattern, error_text)
            ):
                # print("Match 8")
                to_replace = error_text
                self.modified_usfm = self.modified_usfm.replace(
                    to_replace, to_replace[1:]
                )
                changed = True
            # Just a single problematic marker (could be w/o text)
            elif error_text.startswith("\\") and re.match(
                valid_markers_pattern, error_text
            ):
                # print("Match 9")
                start = max(0, error.start_byte - 5)
                end = min(len(self.usfm_bytes), error.end_byte + 5)
                to_replace = self.usfm_bytes[start:end].decode("utf-8")
                repalcement = to_replace.replace(error_text, "")
                # print(to_replace)
                self.modified_usfm = self.modified_usfm.replace(to_replace, repalcement)
                changed = True
            # empty attribute
            elif error_text.strip() == "|":
                # print("Match 10")
                start = max(0, error.start_byte - 5)
                end = min(len(self.usfm_bytes), error.end_byte + 5)
                to_replace = self.usfm_bytes[start:end].decode("utf-8")
                repalcement = to_replace.replace(error_text, "")
                self.modified_usfm = self.modified_usfm.replace(to_replace, repalcement)
                changed = True
            # Stray content in the chapter line
            elif (
                error.parent.type == "chapter"
                and error.prev_sibling.type == "c"
                and "\\" not in error_text
            ):
                # print("Match 10")
                self.modified_usfm = self.modified_usfm.replace(error_text, "")
                changed = True

        # print(f"{changed=}")
        if not changed:
            err_str = self.format_errors()
            self.message = f"Cannot fix these errors:\n\t{err_str}"
            return usfm
        returned_usfm = self.auto_fix_usfm(self.modified_usfm, fixed=True)
        return returned_usfm


book_code_missing_pattern = re.compile(r"\\id[\s\n\r]*\\")
v_without_space_pattern = re.compile(r"(\\v)(\d+)")
c_without_space_pattern = re.compile(r"(\\c)(\d+)")

valid_markers_pattern = re.compile(
    r"(\\id|\\usfm|\\ide|\\ref|\\h|\\toc|\\toca|\\sts|\\rem|"
    + r"\\restore|\\lit|\\iqt|\\iqt|\\imt|\\imte|\\is|\\io|\\ior|\\ior|\\iot|\\ip|\\im|\\ipi|"
    + r"\\imi|\\ili|\\ipq|\\imq|\\ipr|\\ib|\\iq|\\ie|\\iex|\\v|\\va|\\va|\\vp|\\vp|\\c|\\cl|"
    + r"\\ca|\\ca|\\cp|\\cd|\\mt|\\mte|\\ms|\\mr|\\s|\\sr|\\r|\\sp|\\d|\\sd|\\p|\\m|\\po|\\pr|"
    + r"\\cls|\\pmo|\\pm|\\pmc|\\pmr|\\pi|\\mi|\\nb|\\pc|\\ph|\\phi|\\b|\\q|\\qr|\\qc|\\qs|"
    + r"\\qs|\\qa|\\qac|\\qac|\\qm|\\qd|\\lh|\\lf|\\li|\\lim|\\liv|\\lik|\\lik|\\litl|\\litl|"
    + r"\\tr|\\th|\\thr|\\tc|\\tcr|\\f|\\f|\\fe|\\fe|\\ef|\\ef|\\fr|\\fr|\\fq|\\fq|\\fqa|\\fqa|"
    + r"\\fk|\\fk|\\fl|\\fl|\\fw|\\fw|\\fp|\\fp|\\ft|\\ft|\\fdc|\\fdc|\\fv|\\fv|\\fm|\\fm|\\x|"
    + r"\\x|\\xo|\\xo|\\xk|\\xk|\\xq|\\xq|\\xt|\\xt|\\xt|\\xt|\\xta|\\xta|\\xop|\\xop|\\xot|"
    + r"\\xot|\\xnt|\\xnt|\\xdc|\\xdc|\\rq|\\rq|\\add|\\add|\\bk|\\bk|\\dc|\\dc|\\k|\\k|\\nd|"
    + r"\\nd|\\ord|\\ord|\\pn|\\pn|\\png|\\png|\\addpn|\\addpn|\\qt|\\qt|\\sig|\\sig|\\sls|"
    + r"\\sls|\\tl|\\tl|\\wj|\\wj|\\em|\\em|\\bd|\\bd|\\it|\\it|\\bdit|\\bdit|\\no|\\no|\\sc|"
    + r"\\sc|\\sup|\\sup|\\ndx|\\ndx|\\pro|\\pro|\\rb|\\rb|\\w|\\w|\\wg|\\wg|\\wh|\\wh|\\wa|"
    + r"\\wa|\\fig|\\jmp|\\jmp|\\pb|\\z|\\z|\\esb|\\esbe|\\cat|\\cat)(\d|\s|\n|\r|$)"
)



"""Logics for syntax-tree to xml(USX) conversions"""
from lxml import etree
from tree_sitter import QueryCursor
from usfm_grammar.queries import create_queries_as_needed


class USXGenerator:
    """A binding for all methods used in generating USX from Syntax tree"""

    # handled alike by the node_2_usx_generic method
    PARA_STYLE_MARKERS = [
        "ide", "h", "toc", "toca",  # identification
        "imt", "is", "ip", "ipi", "im", "imi", "ipq", "imq",
        "ipr", "iq", "ib", "ili", "iot", "io", "iex", "imte", "ie",  # intro
        "mt", "mte", "cl", "cd", "ms", "mr", "s", "sr", "r", "d", "sp", "sd",  # titles
        "q", "qr", "qc", "qa", "qm", "qd",  # poetry
        "lh", "li", "lf", "lim",  # lists
        "sts", "rem", "lit", "restore",  # comments
        "b",
    ]

    NOTE_MARKERS = ["f", "fe", "ef", "efe", "x", "ex"]
    CHAR_STYLE_MARKERS = [
        "add", "bk", "dc", "ior", "iqt", "k", "litl", "nd", "ord",
        "pn", "png", "qac", "qs", "qt", "rq", "sig", "sls", "tl", "wj",  # Special-text
        "em", "bd", "bdit", "it", "no", "sc", "sup",  # character styling
        "rb", "pro", "w", "wh", "wa", "wg",  # special-features
        "lik", "liv",  # structred list entries
        "jmp", "fr", "ft", "fk", "fq", "fqa", "fl", "fw", "fp", "fv", "fdc",  # footnote-content
        "xo", "xop", "xt", "xta", "xk", "xq", "xot", "xnt", "xdc",  # crossref-content
    ]
    NESTED_CHAR_STYLE_MARKERS = [item + "Nested" for item in CHAR_STYLE_MARKERS]
    DEFAULT_ATTRIB_MAP = {
        "w": "lemma",
        "rb": "gloss",
        "xt": "href",
        "fig": "alt",
        "xt_standalone": "href",
        "xtNested": "href",
        "ref": "loc",
        "milestone": "who",
        "k": "key",
    }
    TABLE_CELL_MARKERS = ["tc", "th", "tcr", "thr", "tcc", "thc"]
    MISC_MARKERS = ["fig", "cat", "esb", "b", "ph", "pi"]
    OTHER_PARA_NESTABLES = [
        "text",
        "footnote",
        "crossref",
        "verseText",
        "v",
        "b",
        "milestone",
        "zNameSpace",
    ]
    MARKER_LISTS = {
        "para_style": PARA_STYLE_MARKERS,
        "note": NOTE_MARKERS,
        "char_style": CHAR_STYLE_MARKERS,
        "nested_char_style": NESTED_CHAR_STYLE_MARKERS,
        "table_cell": TABLE_CELL_MARKERS,
        "misc": MISC_MARKERS,
        "other_para_nestables": OTHER_PARA_NESTABLES,
    }
    MARKER_SETS = {
        "table_cell": set(TABLE_CELL_MARKERS),
        "char_style": set(CHAR_STYLE_MARKERS),
        "nested_char_style": set(NESTED_CHAR_STYLE_MARKERS),
        "other_para_nestables": set(OTHER_PARA_NESTABLES),
        "note_markers": set(NOTE_MARKERS),
        "para_style": set(PARA_STYLE_MARKERS),
        "misc": set(MISC_MARKERS),
    }

    def __init__(self, tree_sitter_language_obj, usfm_bytes, usx_root_element=None):
        """Initialize the class with usfm language and an empty output"""
        self.usfm_language = tree_sitter_language_obj
        self.usfm = usfm_bytes
        if usx_root_element is None:
            self.xml_root_node = etree.Element("usx")
            self.xml_root_node.set("version", "3.1")
        else:
            self.xml_root_node = usx_root_element
        self.parse_state = {
            "book_slug": None,
            "current_chapter": None,
            "prev_verse_parent": None,
            "prev_verse_sid_to_close": None,
            "prev_chapter_sid": None,
            "prev_verse": None,
        }
        self.dispatch_map = self.populate_dispatch_map()
        self.queries = {}

    def get_query(self, name):
        """Get or create a query by name"""
        if name not in self.queries:
            self.queries[name] = self.create_query(name)
        return self.queries[name]

    def create_query(self, name):
        """Create a query by name"""
        return create_queries_as_needed(name, self.usfm_language)

    def node_2_usx_id(self, node, parent_xml_node):
        """build id node in USX"""
        id_cursor = QueryCursor(self.get_query("id"))
        id_captures = id_cursor.captures(node)
        code = self.usfm[id_captures['book-code'][0].start_byte :
                         id_captures['book-code'][0].end_byte].decode("utf-8")\
            if 'book-code' in id_captures else None
        desc = self.usfm[id_captures['desc'][0].start_byte :
                         id_captures['desc'][0].end_byte].decode("utf-8")\
            if 'desc' in id_captures else None

        book_xml_node = etree.SubElement(parent_xml_node, "book")
        book_xml_node.set("code", code)
        book_xml_node.set("style", "id")
        self.parse_state["book_slug"] = code
        if desc is not None and desc.strip() != "":
            book_xml_node.text = desc.strip()

    def node_2_usx_c(self, node, parent_xml_node):
        """Build c, the chapter milestone node in usx"""
        chap_cursor = QueryCursor(self.get_query("chapter"))
        chap_cap = chap_cursor.captures(node)
        chap_num = self.usfm[
            chap_cap['chap-num'][0].start_byte : chap_cap['chap-num'][0].end_byte
        ].decode("utf-8")
        chap_ref = self.parse_state["book_slug"] + " " + chap_num
        self.parse_state["prev_chapter_sid"] = chap_ref
        chap_xml_node = etree.SubElement(parent_xml_node, "chapter")
        chap_xml_node.set("number", chap_num)
        chap_xml_node.set("style", "c")
        chap_xml_node.set("sid", chap_ref)
        self.parse_state["current_chapter"] = chap_num
        alt_num = self.usfm[chap_cap['alt-num'][0].start_byte :
                             chap_cap['alt-num'][0].end_byte].decode("utf-8")\
            if 'alt-num' in chap_cap else None
        pub_num = self.usfm[chap_cap['pub-num'][0].start_byte :
                            chap_cap['pub-num'][0].end_byte].decode("utf-8")\
            if 'pub-num' in chap_cap else None
        if alt_num is not None:
            chap_xml_node.set("altnumber", alt_num.strip())
        if pub_num is not None:
            chap_xml_node.set("pubnumber", pub_num.strip())
        for child in node.children:
            if child.type in ["cl", "cd"]:
                self.node_2_usx(child, parent_xml_node)

    def node_2_usx_chapter(self, node, parent_xml_node):
        """build chapter node in USX"""
        for child in node.children:
            if child.type == "c":
                self.node_2_usx_c(child, parent_xml_node)
            else:
                self.node_2_usx(child, parent_xml_node)

        last_verse = self.parse_state["prev_verse"]
        if last_verse is not None and last_verse.get("eid") is None:
            v_end_xml_node = etree.Element("verse")
            v_end_xml_node.set("eid", last_verse.get("sid"))
            self.parse_state["prev_verse"] = None
            self.parse_state["prev_verse_sid_to_close"] = None
            last_sibbling = parent_xml_node[-1]
            if last_sibbling.tag == "para":
                last_sibbling.append(v_end_xml_node)
            elif last_sibbling.tag == "table":
                rows = list(last_sibbling)
                rows[-1].append(v_end_xml_node)
            else:
                parent_xml_node.append(v_end_xml_node)
        chap_end_xml_node = etree.SubElement(parent_xml_node, "chapter")
        chap_ref = self.parse_state["book_slug"] + " " + self.parse_state["current_chapter"]
        chap_end_xml_node.set("eid", chap_ref)
        self.parse_state["prev_chapter_sid"] = None

    def node_2_usx_verse(self, node, parent_xml_node):
        """build verse node in USX"""
        eid = self.parse_state["prev_verse_sid_to_close"]
        if eid is not None:
            prev_para = self.parse_state["prev_verse_parent"]
            v_end_xml_node = etree.SubElement(prev_para, "verse")
            v_end_xml_node.set("eid", eid)
            self.parse_state["prev_verse_sid_to_close"] = None
        verse_cursor = QueryCursor(self.get_query("verseNumCap"))
        verse_num_cap = verse_cursor.captures(node)
        verse_num = self.usfm[
            verse_num_cap['vnum'][0].start_byte : verse_num_cap['vnum'][0].end_byte
        ].decode("utf-8")
        v_xml_node = etree.SubElement(parent_xml_node, "verse")
        self.parse_state["prev_verse"] = v_xml_node
        alt_num = self.usfm[verse_num_cap['alt'][0].start_byte :
                            verse_num_cap['alt'][0].end_byte].decode("utf-8")\
            if 'alt' in verse_num_cap else None
        pub_num = self.usfm[verse_num_cap['vp'][0].start_byte :
                            verse_num_cap['vp'][0].end_byte].decode("utf-8")\
            if 'vp' in verse_num_cap else None
        if alt_num is not None:
            v_xml_node.set("altnumber", alt_num.strip())
        if pub_num is not None:
            v_xml_node.set("pubnumber", pub_num.strip())
        ref = (
            self.parse_state["book_slug"]
            + " "
            + self.parse_state["current_chapter"]
            + ":"
            + verse_num.strip()
        )
        # ref = self.xml_root_node.findall('.//chapter')[-1].get('sid')+ ":"+ verse_num
        sid = ref.strip()
        v_xml_node.set("number", verse_num.strip())
        v_xml_node.set("style", "v")
        v_xml_node.set("sid", ref.strip())
        self.parse_state["prev_verse_sid_to_close"] = sid

    def node_2_usx_ca_va(self, node, parent_xml_node):
        """Build elements for independant ca and va away from c and v"""
        style = node.type
        char_xml_node = etree.SubElement(parent_xml_node, "char")
        char_xml_node.set("style", style)
        alt_num_cursor = QueryCursor(self.get_query("usjCaVa"))
        alt_num_match = alt_num_cursor.captures(node)[0]
        alt_num = self.usfm[alt_num_match['alt-num'][0].start_byte :
                      alt_num_match['alt-num'][0].end_byte].decode("utf-8").strip()
        char_xml_node.set("altnumber", alt_num)
        char_xml_node.set("closed", "true")

    def node_2_usx_para(self, node, parent_xml_node):
        """build paragraph nodes in USX"""
        if node.children[0].type.endswith("Block"):
            for child in node.children[0].children:
                self.node_2_usx_para(child, parent_xml_node)
        elif node.type == "paragraph":
            para_tag_cursor = QueryCursor(self.get_query("para"))
            para_tag_cap = para_tag_cursor.captures(node)
            para_marker = para_tag_cap['para-marker'][0].type
            if not para_marker.endswith("Block"):
                para_xml_node = etree.SubElement(parent_xml_node, "para")
                para_xml_node.set("style", para_marker)
                for child in para_tag_cap['para-marker'][0].children[1:]:
                    self.node_2_usx(child, para_xml_node)
        elif node.type in ["pi", "ph"]:
            para_marker = (
                self.usfm[node.children[0].start_byte : node.children[0].end_byte]
                .decode("utf-8")
                .replace("\\", "")
                .strip()
            )
            para_xml_node = etree.SubElement(parent_xml_node, "para")
            para_xml_node.set("style", para_marker)
            for child in node.children[1:]:
                self.node_2_usx(child, para_xml_node)

    def node_2_usx_notes(self, node, parent_xml_node):
        """build USX nodes for footnotes and corss-refs"""
        tag_node = node.children[0]
        caller_node = node.children[1]
        note_xml_node = etree.SubElement(parent_xml_node, "note")
        note_xml_node.set(
            "style",
            self.usfm[tag_node.start_byte : tag_node.end_byte]
            .decode("utf-8")
            .replace("\\", "")
            .strip(),
        )
        note_xml_node.set(
            "caller",
            self.usfm[caller_node.start_byte : caller_node.end_byte]
            .decode("utf-8")
            .strip(),
        )
        for child in node.children[2:-1]:
            self.node_2_usx(child, note_xml_node)

    def node_2_usx_char(self, node, parent_xml_node):
        """build USX nodes for character markups, both regular and nested"""
        tag_node = node.children[0]
        closing_node = None
        children_range = len(node.children)
        for i in range(len(node.children)-1, 0, -1):
            if node.children[i].type.startswith("\\") or\
                node.children[i].type == "*" or\
                node.children[i].type.endswith("Tag"):
                children_range -= 1
        char_xml_node = etree.SubElement(parent_xml_node, "char")
        char_xml_node.set(
            "style",
            self.usfm[tag_node.start_byte : tag_node.end_byte]
            .decode("utf-8")
            .replace("\\", "")
            .replace("+", "")
            .strip(),
        )
        if closing_node is None:
            char_xml_node.set("closed", "false")
        else:
            char_xml_node.set("closed", "true")
        for child in node.children[1:children_range]:
            self.node_2_usx(child, char_xml_node)

    def node_2_usx_attrib(self, node, parent_xml_node):
        """add attribute values to USX elements"""
        attrib_name_node = node.children[0]
        attrib_name = (
            self.usfm[attrib_name_node.start_byte : attrib_name_node.end_byte]
            .decode("utf-8")
            .strip()
        )
        if attrib_name == "|":
            parent_type = node.parent.type
            if "Nested" in parent_type:
                parent_type = parent_type.replace("Nested", "")
            attrib_name = self.DEFAULT_ATTRIB_MAP[parent_type]
        if attrib_name == "src":  # for \fig
            attrib_name = "file"

        attrib_cursor = QueryCursor(self.get_query("attribVal"))
        attrib_val_cap = attrib_cursor.captures(node)
        if 'attrib-val' in attrib_val_cap and len(attrib_val_cap['attrib-val']) > 0:
            attrib_value = (
                self.usfm[
                    attrib_val_cap['attrib-val'][0].start_byte :
                    attrib_val_cap['attrib-val'][0].end_byte
                ].decode("utf-8").strip()
            )
        else:
            attrib_value = ""
        parent_xml_node.set(attrib_name, attrib_value)

    def node_2_usx_table(self, node, parent_xml_node):
        """Handle table related components and convert to usx"""
        if node.type == "table":
            table_xml_node = etree.SubElement(parent_xml_node, "table")
            for child in node.children:
                self.node_2_usx(child, table_xml_node)
        elif node.type == "tr":
            row_xml_node = etree.SubElement(parent_xml_node, "row")
            row_xml_node.set("style", "tr")
            for child in node.children[1:]:
                self.node_2_usx(child, row_xml_node)
        elif node.type in self.TABLE_CELL_MARKERS:
            tag_node = node.children[0]
            style = (
                self.usfm[tag_node.start_byte : tag_node.end_byte]
                .decode("utf-8")
                .replace("\\", "")
                .strip()
            )
            cell_xml_node = etree.SubElement(parent_xml_node, "cell")
            cell_xml_node.set("style", style)
            if "r" in style:
                cell_xml_node.set("align", "end")
            elif "tcc" in style or "thc" in style:
                cell_xml_node.set("align", "center")
            else:
                cell_xml_node.set("align", "start")
            for child in node.children[1:]:
                self.node_2_usx(child, cell_xml_node)

    def node_2_usx_milestone(self, node, parent_xml_node):
        """create ms node in USX"""
        ms_name_cursor = QueryCursor(self.get_query("milestone"))
        ms_name_cap = ms_name_cursor.captures(node)['ms-name']
        style = (
            self.usfm[ms_name_cap[0].start_byte : ms_name_cap[0].end_byte]
            .decode("utf-8")
            .replace("\\", "")
            .strip()
        )
        ms_xml_node = etree.SubElement(parent_xml_node, "ms")
        ms_xml_node.set("style", style)
        for child in node.children:
            if child.type.endswith("Attribute"):
                self.node_2_usx(child, ms_xml_node)

    def node_2_usx_special(self, node, parent_xml_node):
        """Build nodes for esb, cat, fig, optbreak in USX"""
        if node.type == "esb":
            style = "esb"
            sidebar_xml_node = etree.SubElement(parent_xml_node, "sidebar")
            sidebar_xml_node.set("style", style)
            for child in node.children[1:-1]:
                self.node_2_usx(child, sidebar_xml_node)
        elif node.type == "cat":
            cat_cursor = QueryCursor(self.get_query("category"))
            cat_cap = cat_cursor.captures(node)['category']
            category = (
                self.usfm[cat_cap[0].start_byte : cat_cap[0].end_byte]
                .decode("utf-8")
                .strip()
            )
            parent_xml_node.set("category", category)
        elif node.type == "fig":
            fig_xml_node = etree.SubElement(parent_xml_node, "figure")
            fig_xml_node.set("style", "fig")
            for child in node.children[1:-1]:
                self.node_2_usx(child, fig_xml_node)
        elif node.type == "ref":
            ref_xml_node = etree.SubElement(parent_xml_node, "ref")
            ref_xml_node.set("style", "ref")
            for child in node.children[1:-1]:
                self.node_2_usx(child, ref_xml_node)

    def node_2_usx_generic(self, node, parent_xml_node):
        """build nodes for para style markers in USX"""
        tag_node = node.children[0]
        style = self.usfm[tag_node.start_byte : tag_node.end_byte].decode("utf-8")
        if style.startswith("\\"):
            style = style.replace("\\", "").strip()
        else:
            style = node.type
        children_range_start = 1
        if len(node.children) > 1 and node.children[1].type.startswith("numbered"):
            num_node = node.children[1]
            num = self.usfm[num_node.start_byte : num_node.end_byte].decode("utf-8")
            style += num
            children_range_start = 2
        para_xml_node = etree.SubElement(parent_xml_node, "para")
        para_xml_node.set("style", style)
        for child in node.children[children_range_start:]:
            # self.node_2_usx(child, para_xml_node)
            if any(
                child.type in marker_set
                for marker_set in [
                    self.CHAR_STYLE_MARKERS,
                    self.NESTED_CHAR_STYLE_MARKERS,
                    self.OTHER_PARA_NESTABLES,
                ]
            ):
                self.node_2_usx(child, para_xml_node)
            else:
                self.node_2_usx(child, parent_xml_node)

    def populate_dispatch_map(self):
        """Create a dispatch map for node type to handler functions"""
        dispatch_map = {}

        def add_handlers(markers, handler):
            for marker in markers:
                if isinstance(marker, (list, set)):
                    for mark in marker:
                        dispatch_map[mark] = getattr(self, handler.__name__)
                else:
                    dispatch_map[marker] = getattr(self, handler.__name__)

        # Add basic handlers
        dispatch_map["text"] = self.push_text_node
        dispatch_map["verseText"] = self.handle_verse_text
        dispatch_map["v"] = self.node_2_usx_verse
        dispatch_map["id"] = self.node_2_usx_id
        dispatch_map["chapter"] = self.node_2_usx_chapter
        dispatch_map["usfm"] = lambda *_: None  # noop

        # Add handlers for different marker types
        add_handlers(["paragraph", "q", "w"], self.node_2_usx_para)
        add_handlers(["cl", "cp", "vp"], self.node_2_usx_generic)
        add_handlers(["ca", "va"], self.node_2_usx_ca_va)
        add_handlers(["table", "tr"], self.node_2_usx_table)
        add_handlers(["milestone", "zNameSpace"], self.node_2_usx_milestone)
        add_handlers(["esb", "cat", "fig", "ref"], self.node_2_usx_special)
        add_handlers(self.NOTE_MARKERS, self.node_2_usx_notes)
        add_handlers(
            self.CHAR_STYLE_MARKERS
            + self.NESTED_CHAR_STYLE_MARKERS
            + ["xt_standalone"],
            self.node_2_usx_char,
        )
        add_handlers(self.TABLE_CELL_MARKERS, self.node_2_usx_table)

        # Add paragraph style markers
        for marker in self.PARA_STYLE_MARKERS:
            if marker != "usfm":
                dispatch_map[marker] = self.node_2_usx_generic

        return dispatch_map

    def handle_verse_text(self, node, parent_xml_node):
        """Handle verse text content"""
        for child in node.children:
            self.node_2_usx(child, parent_xml_node)
        self.parse_state["prev_verse_parent"] = parent_xml_node

    def push_text_node(self, node, parent_xml_node):
        """Add text node to parent's content"""
        text_val = (
            self.usfm[node.start_byte : node.end_byte].decode("utf-8").replace("~", " ")
        )
        if text_val:
            if len(parent_xml_node) > 0:
                # If there are already children, append text to the tail of the last child
                last_child = parent_xml_node[-1]
                if last_child.tail is None:
                    last_child.tail = text_val
                else:
                    last_child.tail += text_val
            else:
                # If no children, set the text of the parent directly
                if parent_xml_node.text is None:
                    parent_xml_node.text = text_val
                else:
                    parent_xml_node.text += text_val

    def node_2_usx(self, node, parent_xml_node):  # pylint: disable= too-many-branches
        """check each node and based on the type convert to corresponding xml element"""
        # print("working with node: ", node, "\n")
        if not hasattr(node, "type"):
            return
        node_type = node.type.replace("\\", "") if node.type else ""
        handler = self.dispatch_map.get(node_type)
        if handler:
            handler(node, parent_xml_node)
        elif node_type.endswith("Attribute"):
            self.node_2_usx_attrib(node, parent_xml_node)
        elif node_type.strip() in ["", "|"]:
            pass  # skip white space nodes
        elif hasattr(node, "children") and len(node.children) > 0:
            for child in node.children:
                self.node_2_usx(child, parent_xml_node)
        # else:
        #     raise Exception("Encountered unknown element ", str(node))



"""Entry point for the package, when invoked from command line"""

import argparse
import json
import sys
import csv
from lxml import etree

from usfm_grammar import USFMParser, Filter, Format, ORIGINAL_VREF
from usfm_grammar import __version__
from usfm_grammar.errors import ParameterError

all_markers = []
for member in Filter:
    all_markers += member.value


def handle_input_file(arg_parser):
    """If initialsing with USFM or USJ?"""
    infile = arg_parser.parse_args().infile
    input_format = arg_parser.parse_args().in_format
    with open(infile, "r", encoding="utf-8") as usfm_file:
        file_content = usfm_file.read()

    if input_format == Format.JSON or infile.split(".")[-1].lower() in ["json", "usj"]:
        usj_obj = json.loads(file_content)
        my_parser = USFMParser(from_usj=usj_obj)
    elif input_format == Format.USX or infile.split(".")[-1].lower() in ["xml", "usx"]:
        usx_obj = etree.fromstring(file_content)
        my_parser = USFMParser(from_usx=usx_obj)
    elif input_format == Format.USFM:
        my_parser = USFMParser(file_content)
    elif input_format == Format.BIBLENLP:
        texts = file_content.rstrip().split("\n")
        vref_file = arg_parser.parse_args().vref
        bookcode = arg_parser.parse_args().bookcode
        if vref_file is None:
            refs = ORIGINAL_VREF
        else:
            with open(vref_file, "r", encoding="utf-8") as vrf:
                refs = vrf.read().rstrip().split("\n")
        obj = {"vref": refs, "text": texts}
        my_parser = USFMParser(from_biblenlp=obj, book_code=bookcode)
    else:
        raise ParameterError("Un-recognized input_format!")
    return my_parser


def handle_include_exclude_options(arg_parser):
    """Process list of markers and ENUMs"""
    exclude_markers = arg_parser.parse_args().exclude_markers
    include_markers = arg_parser.parse_args().include_markers

    filter_names = [member.name for member in Filter]
    if exclude_markers is None:
        updated_exclude_markers = None
    else:
        updated_exclude_markers = []
        for itm in exclude_markers:
            if itm.upper() in filter_names:
                updated_exclude_markers += Filter[itm.upper()]
            else:
                updated_exclude_markers.append(itm.lower().replace("\\", ""))
    if include_markers is None:
        updated_include_markers = None
    else:
        updated_include_markers = []
        for itm in include_markers:
            if itm.upper() in filter_names:
                updated_include_markers += Filter[itm.upper()]
            else:
                updated_include_markers.append(itm.lower().replace("\\", ""))
    return updated_exclude_markers, updated_include_markers


def main():  # pylint: disable=too-many-locals
    """handles the command line requests"""
    arg_parser = argparse.ArgumentParser(
        description="Uses the tree-sitter-usfm grammar to parse and convert USFM to "
        + "Syntax-tree, JSON, CSV, USX etc."
    )
    arg_parser.add_argument("infile", type=str, help="input usfm or usj file")

    arg_parser.add_argument(
        "--in_format",
        type=str,
        help="input file format",
        choices=[
            Format.USFM.value,
            Format.JSON.value,
            Format.USX.value,
            Format.BIBLENLP.value,
        ],
        default=Format.USFM.value,
    )
    arg_parser.add_argument(
        "--out_format",
        type=str,
        help="output format",
        choices=[itm.value for itm in Format],
        default=Format.JSON.value,
    )
    arg_parser.add_argument(
        "--include_markers",
        type=str,
        help="the list of of contents to be included",
        choices=[itm.name.lower() for itm in Filter] + all_markers,
        action="append",
    )
    arg_parser.add_argument(
        "--exclude_markers",
        type=str,
        help="the list of of contents to be included",
        choices=[itm.name.lower() for itm in Filter] + all_markers,
        action="append",
    )
    arg_parser.add_argument(
        "--csv_col_sep",
        type=str,
        help="column separator or delimiter. Only useful with format=table.",
        default="\t",
    )
    arg_parser.add_argument(
        "--csv_row_sep",
        type=str,
        help="row separator or delimiter. Only useful with format=table.",
        default="\n",
    )
    arg_parser.add_argument(
        "--ignore_errors",
        help="to get some output from successfully parsed portions",
        action="store_true",
    )
    arg_parser.add_argument(
        "--combine_text",
        help="to be used along with exclude_markers or include_markers, "
        + "to concatinate the consecutive text snippets, "
        + "from different components, or not",
        action="store_true",
    )
    arg_parser.add_argument(
        "--vref",
        help="path to the vref file containing line by line verse reference"
        + " for biblenlp input file",
        default=None,
    )
    arg_parser.add_argument(
        "--bookcode",
        help="book to be exported from biblenlp to usfm format",
        default=None,
    )
    arg_parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )

    my_parser = handle_input_file(arg_parser)

    if my_parser.errors and not arg_parser.parse_args().ignore_errors:
        err_str = "\n\t".join([":".join(err) for err in my_parser.errors])
        print(f"Errors present:\n\t{err_str}")
        sys.exit(1)

    exclude_markers, include_markers = handle_include_exclude_options(arg_parser)

    output_format = arg_parser.parse_args().out_format

    ignore_errors = arg_parser.parse_args().ignore_errors

    match output_format:
        case Format.JSON:
            dict_output = my_parser.to_usj(
                exclude_markers=exclude_markers,
                include_markers=include_markers,
                ignore_errors=ignore_errors,
            )
            print(json.dumps(dict_output, indent=4, ensure_ascii=False))
        case Format.CSV:
            table_output = my_parser.to_list(
                exclude_markers=exclude_markers,
                include_markers=include_markers,
                ignore_errors=ignore_errors,
            )
            outfile = sys.stdout
            writer = csv.writer(
                outfile,
                delimiter=arg_parser.parse_args().csv_col_sep,
                lineterminator=arg_parser.parse_args().csv_row_sep,
            )
            writer.writerows(table_output)
        case Format.USX:
            xmlstr = etree.tostring(
                my_parser.to_usx(ignore_errors=ignore_errors),
                encoding="unicode",
                pretty_print=True,
            )
            print(xmlstr)
        case Format.MD:
            print(my_parser.to_markdown())
        case Format.ST:
            print(my_parser.to_syntax_tree(ignore_errors=ignore_errors))
        case Format.USFM:
            print(my_parser.usfm)
        case Format.BIBLENLP:
            infile = arg_parser.parse_args().infile
            file_parts = infile.split(".")
            file_parts[-1] = "txt"
            file_parts[-2] += "_biblenlp"
            outfile_name = ".".join(file_parts)
            outfile_name2 = outfile_name.replace("_biblenlp.txt", "_biblenlp_vref.txt")
            bible_nlp_dict = my_parser.to_biblenlp_format(ignore_errors=ignore_errors)
            with open(outfile_name, "w", encoding="utf-8") as out1:
                out1.writelines(f"{line}\n" for line in bible_nlp_dict["text"])
            with open(outfile_name2, "w", encoding="utf-8") as out2:
                out2.writelines(f"{line}\n" for line in bible_nlp_dict["vref"])
            print(f"Outputs written to {outfile_name} and {outfile_name2}.")
        case _:
            raise ParameterError(f"Un-recognized output format:{output_format}!")


if __name__ == "__main__":
    main()



"""JSON schema definition included within the package code"""

usj_schema = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "$id": "https://github.com/usfm-bible/tcdocs/blob/main/grammar/usj.js",
    "title": "Unified Scripture JSON",
    "description": "The JSON varient of USFM and USX data models",
    "type": "object",
    "$defs": {
        "markerObject": {
            "type": "object",
            "properties": {
                "type": {
                    "description": "The kind/category of node or element this is,"
                    + "corresponding the USFM marker and USX node",
                    "type": "string",
                },
                "marker": {
                    "description": "The corresponding marker in USFM or style in USX",
                    "type": "string",
                },
                "content": {
                    "type": "array",
                    "items": {
                        "anyOf": [{"type": "string"}, {"$ref": "#/$defs/markerObject"}]
                    },
                },
                "sid": {
                    "description": (
                        "Indicates the Book-chapter-verse value in the "
                        "paragraph based structure"
                    ),
                    "type": "string",
                },
                "number": {
                    "description": "Chapter number or verse number",
                    "type": "string",
                },
                "code": {
                    "description": "The 3-letter book code in id element",
                    "pattern": "^[0-9A-Z]{3}$",
                    "type": "string",
                },
                "altnumber": {
                    "description": "Alternate chapter number or verse number",
                    "type": "string",
                },
                "pubnumber": {
                    "description": "Published character of chapter or verse",
                    "type": "string",
                },
                "caller": {
                    "description": "Caller character for footnotes and cross-refs",
                    "type": "string",
                },
                "align": {"description": "Alignment of table cells", "type": "string"},
                "category": {
                    "description": "Category of extended study bible sections",
                    "type": "string",
                },
            },
            "required": ["type"],
        }
    },
    "properties": {
        "type": {
            "description": "The kind of node/element/marker this is",
            "type": "string",
        },
        "version": {"description": "The USJ spec version", "type": "string"},
        "content": {
            "description": "The JSON representation of scripture contents from USFM/USX",
            "type": "array",
            "items": {"anyOf": [{"type": "string"}, {"$ref": "#/$defs/markerObject"}]},
        },
    },
    "required": ["type", "version", "content"],
}



"""Logics for USJ to list(table) and Bible NLP format conversions"""


class ListGenerator:
    """Combines the methods used for List generattion from USJ"""

    def __init__(self):
        """Variables shared by functions"""
        self.book = ""
        self.current_chapter = ""
        self.current_verse = ""
        self.list = [["Book", "Chapter", "Verse", "Text", "Type", "Marker"]]
        self.bible_nlp_format = {"text": [], "vref": []}
        self.prev_chapter = ""
        self.prev_verse = ""

    def usj_to_list_id(self, obj):
        """update book code"""
        self.book = obj["code"]

    def usj_to_list_c(self, obj):
        """Update current chapter"""
        self.current_chapter = obj["number"]
        self.current_verse = ""

    def usj_to_list_v(self, obj):
        """Update current verse"""
        self.current_verse = obj["number"]

    def usj_to_list(self, obj, exclude_markers=None, include_markers=None):  # pylint: disable=too-many-branches
        """Traverse the USJ dict and build the table in self.list"""
        if obj["type"] == "book":
            self.usj_to_list_id(obj)
            if (exclude_markers and "id" in exclude_markers) or (
                include_markers and "id" not in include_markers
            ):
                return
        elif obj["type"] == "chapter":
            self.usj_to_list_c(obj)
        elif obj["type"] == "verse":
            self.usj_to_list_v(obj)
        marker_type = obj["type"]
        marker_name = obj["marker"] if "marker" in obj else ""
        if marker_type == "USJ":
            # This would occur if the JSON got flatttened after removing paragraph markers
            marker_type = ""
        if "content" in obj:
            for item in obj["content"]:
                if isinstance(item, str):
                    if exclude_markers and "text" in exclude_markers:
                        item = ""
                    self.list.append(
                        [
                            self.book,
                            self.current_chapter,
                            self.current_verse,
                            item,
                            marker_type,
                            marker_name,
                        ]
                    )
                else:
                    self.usj_to_list(item, exclude_markers, include_markers)
        if "content" not in obj or len(obj["content"]) == 0:
            if not exclude_markers and not include_markers:
                self.list.append(
                    [
                        self.book,
                        self.current_chapter,
                        self.current_verse,
                        "",
                        marker_type,
                        marker_name,
                    ]
                )
            elif (exclude_markers and marker_name not in exclude_markers) or (
                include_markers and marker_name in include_markers
            ):
                self.list.append(
                    [
                        self.book,
                        self.current_chapter,
                        self.current_verse,
                        "",
                        marker_type,
                        marker_name,
                    ]
                )

    def usj_to_biblenlp_format(self, obj):
        """Traverse the USJ dict and build a dict for bible nlp format, in self.bible_nlp_format"""
        if obj["type"] == "book":
            self.usj_to_list_id(obj)
        elif obj["type"] == "chapter":
            self.usj_to_list_c(obj)
        elif obj["type"] == "verse":
            self.usj_to_list_v(obj)
        marker_type = obj["type"]
        if marker_type == "USJ":
            # This would occur if the JSON got flatttened after removing paragraph markers
            marker_type = ""
        if marker_type != "book" and "content" in obj:
            for item in obj["content"]:
                if isinstance(item, str):
                    if (
                        self.current_chapter == self.prev_chapter
                        and self.current_verse == self.prev_verse
                    ):
                        self.bible_nlp_format["text"][-1] += (
                            " " + item.replace("\n", " ").strip()
                        )
                    else:
                        vref = (
                            f"{self.book} {self.current_chapter}:{self.current_verse}"
                        )
                        self.bible_nlp_format["text"].append(
                            item.replace("\n", " ").strip()
                        )
                        self.bible_nlp_format["vref"].append(vref)
                        self.prev_chapter = self.current_chapter
                        self.prev_verse = self.current_verse
                else:
                    self.usj_to_biblenlp_format(item)



"""Filtering implementations
1. Removing unwanted markers from USJ, provided the list
2. Inlcuding only the given list of markers in USJ
"""

import re

MARKERS_WITH_DISCARDABLE_CONTENTS = [
    "ide", "usfm", "h", "toc", "toca",  # identification
    "imt", "is", "ip", "ipi", "im", "imi", "ipq", "imq", "ipr", "iq", "ib",
    "ili", "iot", "io", "iex", "imte", "ie",  # intro
    "mt", "mte", "cl", "cd", "ms", "mr", "s", "sr", "r", "d", "sp", "sd",  # titles
    "sts", "rem", "lit", "restore",  # comments
    "f", "fe", "ef", "efe", "x", "ex",  # NOTE_MARKERS
    "fr", "ft", "fk", "fq", "fqa", "fl", "fw", "fp", "fv", "fdc",  # footnote-content
    "xo", "xop", "xt", "xta", "xk", "xq", "xot", "xnt", "xdc",  # crossref-content
    "jmp", "fig", "cat", "esb", "b",
]

trailing_num_pattern = re.compile(r"\d+$")
punct_pattern_no_space_before = re.compile(r"^[,.\-—/;:!?@$%^)}\]>”»]")
punct_pattern_no_space_after = re.compile(r"[\-—/`@^&({[<“«]$")
# both lists exculde ', ", *, &, #, ~, |, +, _, =, \


def combine_consequtive_text_contents(contents_list):
    """After filtering, if content endups with text items next to each other, concatinate them"""
    text_combined_contents = []
    text_contents = ""
    for item in contents_list:
        if isinstance(item, str):
            if not (
                text_contents.endswith(" ")
                or item.startswith(" ")
                or text_contents == ""
                or re.match(punct_pattern_no_space_before, item)
                or re.match(punct_pattern_no_space_after, text_contents)
            ):
                text_contents += " "
            text_contents += item
        else:
            if text_contents != "":
                text_combined_contents.append(text_contents)
                text_contents = ""
            text_combined_contents.append(item)
    if text_contents != "":
        text_combined_contents.append(text_contents)
    return text_combined_contents


def exclude_markers_in_usj(
    input_usj, exclude_markers: list, combine_texts=True, excluded_parent=False
):
    """Removing unwanted markers from USJ, provided the list"""
    if isinstance(input_usj, str):
        if excluded_parent and "text-in-excluded-parent" in exclude_markers:
            return []
        return [input_usj]
    cleaned_kids = []
    exclude_markers = [
        re.sub(trailing_num_pattern, "", item) for item in exclude_markers
    ]
    this_marker = input_usj["marker"] if "marker" in input_usj else ""
    this_marker = re.sub(trailing_num_pattern, "", this_marker)
    this_marker_needed = True
    excluded_parent = (
        False  # used to check if its text is needed or not, in the subsequent call
    )
    inner_content_needed = True
    if this_marker in exclude_markers:
        this_marker_needed = False
        excluded_parent = True
        if this_marker in MARKERS_WITH_DISCARDABLE_CONTENTS:
            inner_content_needed = False
    if (this_marker_needed or inner_content_needed) and "content" in input_usj:
        for item in input_usj["content"]:
            cleaned_up_kid = exclude_markers_in_usj(
                item, exclude_markers, combine_texts, excluded_parent
            )
            if isinstance(cleaned_up_kid, list):
                cleaned_kids += cleaned_up_kid
            else:
                cleaned_kids.append(cleaned_up_kid)
        if combine_texts:
            cleaned_kids = combine_consequtive_text_contents(cleaned_kids)
    if this_marker_needed:
        cleaned_usj = input_usj.copy()
        cleaned_usj["content"] = cleaned_kids
        return cleaned_usj
    if inner_content_needed:
        return cleaned_kids
    return []


def include_markers_in_usj(
    input_usj, include_markers: list, combine_texts=True, excluded_parent=False
):
    """keeping only chosen markers in USJ, as per provided the list"""
    if isinstance(input_usj, str):
        if excluded_parent and "text-in-excluded-parent" not in include_markers:
            return []
        return [input_usj]
    cleaned_kids = []
    include_markers = [
        re.sub(trailing_num_pattern, "", item) for item in include_markers
    ]
    this_marker = input_usj["marker"] if "marker" in input_usj else ""
    this_marker = re.sub(trailing_num_pattern, "", this_marker)
    this_marker_needed = True
    excluded_parent = (
        False  # used to check if its text is needed or not in the subsequent call
    )
    inner_content_needed = True
    if this_marker not in include_markers + [""]:
        this_marker_needed = False
        excluded_parent = True
        if this_marker in MARKERS_WITH_DISCARDABLE_CONTENTS:
            inner_content_needed = False
    if (this_marker_needed or inner_content_needed) and "content" in input_usj:
        for item in input_usj["content"]:
            cleaned_up_kid = include_markers_in_usj(
                item, include_markers, combine_texts, excluded_parent
            )
            if isinstance(cleaned_up_kid, list):
                cleaned_kids += cleaned_up_kid
            else:
                cleaned_kids.append(cleaned_up_kid)
        if combine_texts:
            cleaned_kids = combine_consequtive_text_contents(cleaned_kids)
    if this_marker_needed:
        cleaned_usj = input_usj.copy()
        cleaned_usj["content"] = cleaned_kids
        return cleaned_usj
    if inner_content_needed:
        return cleaned_kids
    return []
